import logging
import importlib
from deltacat.compute.compactor_v2.model.merge_input import MergeInput
import numpy as np
import pyarrow as pa
import ray
import itertools
import time
import pyarrow.compute as pc
from deltacat.utils.pyarrow import MAX_INT_BYTES
import deltacat.compute.compactor_v2.utils.merge as merge_utils
from uuid import uuid4
from deltacat import logs
from typing import Callable, Iterator, List, Optional, Tuple, Set
from deltacat.compute.compactor_v2.model.merge_result import MergeResult
from deltacat.compute.compactor_v2.model.merge_file_group import MergeFileGroup
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.compute.compactor import RoundCompletionInfo, DeltaFileEnvelope
from deltacat.utils.common import ReadKwargsProvider
from contextlib import nullcontext
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics, failure_metric, success_metric
from deltacat.utils.resources import (
    get_current_process_peak_memory_usage_in_bytes,
    ProcessUtilizationOverTimeRange,
)
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    generate_pk_hash_column,
    pk_digest_to_hash_bucket_index,
)
from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    Partition,
    metastore,
)
from deltacat.storage.model.manifest import Manifest
from deltacat.compute.compactor_v2.utils.dedupe import drop_duplicates
from deltacat.constants import BYTES_PER_GIBIBYTE
from deltacat.compute.compactor_v2.constants import (
    MERGE_TIME_IN_SECONDS,
    MERGE_SUCCESS_COUNT,
    MERGE_FAILURE_COUNT,
    BUCKETING_SPEC_COMPLIANCE_PROFILE,
    BUCKETING_SPEC_COMPLIANCE_ASSERT,
    BUCKETING_SPEC_COMPLIANCE_PRINT_LOG,
)
from deltacat.exceptions import (
    categorize_errors,
)

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


_EXISTING_VARIANT_LOG_PREFIX = "Existing variant "
_INCREMENTAL_TABLE_LOG_PREFIX = "Incremental table "


def _append_delta_type_column(table: pa.Table, value: np.bool_):
    return table.append_column(
        sc._DELTA_TYPE_COLUMN_FIELD,
        pa.array(np.repeat(value, len(table)), sc._DELTA_TYPE_COLUMN_TYPE),
    )


def _drop_delta_type_rows(table: pa.Table, delta_type: DeltaType) -> pa.Table:
    if sc._DELTA_TYPE_COLUMN_NAME not in table.column_names:
        return table

    delta_type_value = sc.delta_type_to_field(delta_type)

    result = table.filter(
        pc.not_equal(table[sc._DELTA_TYPE_COLUMN_NAME], delta_type_value)
    )

    return result.drop([sc._DELTA_TYPE_COLUMN_NAME])


def _build_incremental_table(
    df_envelopes: List[DeltaFileEnvelope],
) -> pa.Table:

    hb_tables = []
    # sort by delta file stream position now instead of sorting every row later
    is_delete = False
    for df_envelope in df_envelopes:
        # Allow APPEND, UPSERT, and DELETE delta types
        assert df_envelope.delta_type in (
            DeltaType.APPEND,
            DeltaType.UPSERT,
            DeltaType.DELETE,
        ), "Only APPEND, UPSERT, and DELETE delta types are supported"
        if df_envelope.delta_type == DeltaType.DELETE:
            is_delete = True

    for df_envelope in df_envelopes:
        table = df_envelope.table
        if is_delete:
            table = _append_delta_type_column(
                table, np.bool_(sc.delta_type_to_field(df_envelope.delta_type))
            )

        hb_tables.append(table)
    result = _concat_or_coerce_tables(hb_tables)
    return result


def _concat_or_coerce_tables(all_tables: List[pa.Table]) -> pa.Table:
    try:
        return pa.concat_tables(all_tables)
    except pa.ArrowInvalid:
        # Fallback path: schema evolution needed - try PyArrow's built-in unification
        if all_tables:
            try:
                return pa.concat_tables(
                    all_tables, promote_options="permissive", unify_schemas=True
                )
            except (pa.ArrowInvalid, TypeError, pa.ArrowNotImplementedError):
                # If PyArrow unification fails, re-raise the original error
                raise
        else:
            # Empty table list - should not happen but handle gracefully
            raise RuntimeError("Expected at least one table to merge, but found none.")


def _merge_tables(
    table: pa.Table,
    primary_keys: List[str],
    can_drop_duplicates: bool,
    hb_index: int,
    num_buckets: int,
    original_fields: Set[str],
    compacted_table: Optional[pa.Table] = None,
) -> pa.Table:
    """
    Merges the table with compacted table dropping duplicates where necessary.

    This method ensures the appropriate deltas of types [UPSERT] are correctly
    appended to the table.
    """

    all_tables = []
    incremental_idx = 0

    if compacted_table:
        incremental_idx = 1
        all_tables.append(compacted_table)

    all_tables.append(table)

    check_bucketing_spec = BUCKETING_SPEC_COMPLIANCE_PROFILE in [
        BUCKETING_SPEC_COMPLIANCE_PRINT_LOG,
        BUCKETING_SPEC_COMPLIANCE_ASSERT,
    ]

    if primary_keys and check_bucketing_spec:
        _validate_bucketing_spec_compliance(
            table=all_tables[incremental_idx],
            num_buckets=num_buckets,
            primary_keys=primary_keys,
            hb_index=hb_index,
            log_prefix=_INCREMENTAL_TABLE_LOG_PREFIX,
        )

    if not primary_keys or not can_drop_duplicates:
        logger.info(
            f"Not dropping duplicates for primary keys={primary_keys} "
            f"and can_drop_duplicates={can_drop_duplicates}"
        )
        all_tables[incremental_idx] = _drop_delta_type_rows(
            all_tables[incremental_idx], DeltaType.DELETE
        )
        # we need not drop duplicates
        return _concat_or_coerce_tables(all_tables)

    all_tables = generate_pk_hash_column(all_tables, primary_keys=primary_keys)

    result_table_list = []

    incremental_table = drop_duplicates(
        all_tables[incremental_idx], on=sc._PK_HASH_STRING_COLUMN_NAME
    )

    # Always drop DELETE rows from incremental table
    incremental_table = _drop_delta_type_rows(incremental_table, DeltaType.DELETE)

    # Default to using incremental records as-is, override only if merging is needed
    incremental_data = incremental_table

    if compacted_table:
        compacted_table = all_tables[0]

        compacted_pk_hash_str = compacted_table[sc._PK_HASH_STRING_COLUMN_NAME]
        incremental_pk_hash_str = incremental_table[sc._PK_HASH_STRING_COLUMN_NAME]

        logger.info(
            f"Size of compacted pk hash={compacted_pk_hash_str.nbytes} "
            f"and incremental pk hash={incremental_pk_hash_str.nbytes}."
        )

        if (
            compacted_table[sc._PK_HASH_STRING_COLUMN_NAME].nbytes >= MAX_INT_BYTES
            or incremental_table[sc._PK_HASH_STRING_COLUMN_NAME].nbytes >= MAX_INT_BYTES
        ):
            logger.info("Casting compacted and incremental pk hash to large_string...")
            # is_in combines the chunks of the chunked array passed which can cause
            # ArrowCapacityError if the total size of string array is over 2GB.
            # Using a large_string would resolve this issue.
            # The cast here should be zero-copy in most cases.
            compacted_pk_hash_str = pc.cast(compacted_pk_hash_str, pa.large_string())
            incremental_pk_hash_str = pc.cast(
                incremental_pk_hash_str, pa.large_string()
            )

        records_to_update = pc.is_in(
            compacted_pk_hash_str,
            incremental_pk_hash_str,
        )

        records_to_keep = pc.invert(records_to_update)

        # Keep records that don't have updates
        result_table_list.append(compacted_table.filter(records_to_keep))

        # Override default if merging is needed
        if pc.sum(records_to_update).as_py() > 0:  # There are records to update
            old_records_to_update = compacted_table.filter(records_to_update)
            # Perform partial UPSERT: merge old and new records field by field
            incremental_data = _merge_records_partially(
                old_records=old_records_to_update,
                new_records=incremental_table,
                original_fields=original_fields,
            )

    # Add the determined incremental data
    result_table_list.append(incremental_data)

    final_table = _concat_or_coerce_tables(result_table_list)
    final_table = final_table.drop([sc._PK_HASH_STRING_COLUMN_NAME])

    return final_table


def _merge_records_partially(
    old_records: pa.Table, new_records: pa.Table, original_fields: Set[str]
) -> pa.Table:
    """
    Merge records field by field for partial UPSERT behavior. Fills missing
    fields in new_records with values from old_records.

    Args:
        old_records: Records from the compacted table that need updates
        new_records: New records with potential partial field updates

    Returns:
        Table with merged records where missing fields preserve old values
    """
    # Get field sets (excluding hash column which is used for joining)
    old_fields = set(old_records.column_names) - {sc._PK_HASH_STRING_COLUMN_NAME}
    new_fields = set(new_records.column_names) - {sc._PK_HASH_STRING_COLUMN_NAME}

    # Find fields that are missing from new_records but exist in old_records
    missing_fields = old_fields - new_fields

    # Find fields that were auto-added by schema coercion (missing from original user data)
    # These should be treated as missing fields and filled from old_records
    auto_added_null_fields = set()

    # Use definitive information about which fields were originally provided
    # Any field that exists in both tables but was NOT in the original user data
    # should be treated as auto-added by schema coercion
    for field_name in old_fields & new_fields:  # Fields that exist in both
        if field_name not in original_fields:
            auto_added_null_fields.add(field_name)

    # Combine missing fields with auto-added null fields
    fields_to_fill = missing_fields | auto_added_null_fields

    # Start with new_records and add missing fields from old_records
    result_columns = {}

    # Copy all existing columns from new_records
    for column_name in new_records.column_names:
        result_columns[column_name] = new_records[column_name]

    # Fill in missing/auto-added null fields with values from old_records
    for field_name in fields_to_fill:
        # For missing fields, use the old values entirely
        result_columns[field_name] = old_records[field_name]

    # Create the enhanced new_records table with all fields filled
    enhanced_new_records = pa.table(result_columns)

    # Now we can return the enhanced table - it has all the fields with proper values
    # Missing fields are filled with old values, explicitly null fields remain null
    return enhanced_new_records


def _validate_bucketing_spec_compliance(
    table: pa.Table,
    num_buckets: int,
    hb_index: int,
    primary_keys: List[str],
    rci: Optional[RoundCompletionInfo] = None,
    log_prefix=None,
) -> None:
    if rci is not None:
        message_prefix = f"{log_prefix}{rci.compacted_delta_locator.namespace}.{rci.compacted_delta_locator.table_name}.{rci.compacted_delta_locator.table_version}.{rci.compacted_delta_locator.partition_id}.{rci.compacted_delta_locator.partition_values}"
    else:
        message_prefix = f"{log_prefix}"
    pki_table = generate_pk_hash_column(
        [table], primary_keys=primary_keys, requires_hash=True
    )[0]
    is_not_compliant: bool = False
    for index, hash_value in enumerate(sc.pk_hash_string_column_np(pki_table)):
        hash_bucket: int = pk_digest_to_hash_bucket_index(hash_value, num_buckets)
        if hash_bucket != hb_index:
            is_not_compliant = True
            logger.info(
                f"{message_prefix} has non-compliant bucketing spec at index: {index} "
                f"Expected hash bucket is {hb_index} but found {hash_bucket}."
            )
            if BUCKETING_SPEC_COMPLIANCE_PROFILE == BUCKETING_SPEC_COMPLIANCE_ASSERT:
                raise AssertionError(
                    f"Hash bucket drift detected at index: {index}. Expected hash bucket index"
                    f" to be {hb_index} but found {hash_bucket}"
                )
            # No further checks necessary
            break
    if not is_not_compliant:
        logger.debug(
            f"{message_prefix} has compliant bucketing spec for hb_index: {hb_index}"
        )


def _download_compacted_table(
    hb_index: int,
    rci: RoundCompletionInfo,
    primary_keys: List[str],
    all_column_names: List[str],
    compacted_delta_manifest: Optional[Manifest] = None,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    deltacat_storage: metastore = metastore,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> pa.Table:
    tables = []
    hb_index_to_indices = rci.hb_index_to_entry_range

    if str(hb_index) not in hb_index_to_indices:
        return None
    indices = hb_index_to_indices[str(hb_index)]
    assert (
        indices is not None and len(indices) == 2
    ), "indices should not be none and contains exactly two elements"
    for offset in range(indices[1] - indices[0]):
        table = deltacat_storage.download_delta_manifest_entry(
            Delta.of(
                rci.compacted_delta_locator,
                DeltaType.APPEND,
                compacted_delta_manifest.meta,
                None,
                compacted_delta_manifest,
            ),
            entry_index=(indices[0] + offset),
            file_reader_kwargs_provider=read_kwargs_provider,
            all_column_names=all_column_names,
            **deltacat_storage_kwargs,
        )

        tables.append(table)

    compacted_table = pa.concat_tables(tables)
    check_bucketing_spec = BUCKETING_SPEC_COMPLIANCE_PROFILE in [
        BUCKETING_SPEC_COMPLIANCE_PRINT_LOG,
        BUCKETING_SPEC_COMPLIANCE_ASSERT,
    ]

    logger.debug(
        f"Value of BUCKETING_SPEC_COMPLIANCE_PROFILE, check_bucketing_spec:"
        f" {BUCKETING_SPEC_COMPLIANCE_PROFILE}, {check_bucketing_spec}"
    )

    # Bucketing spec compliance isn't required without primary keys
    if primary_keys and check_bucketing_spec:
        _validate_bucketing_spec_compliance(
            compacted_table,
            rci.hash_bucket_count,
            hb_index,
            primary_keys,
            rci=rci,
            log_prefix=_EXISTING_VARIANT_LOG_PREFIX,
        )
    return compacted_table


def _copy_all_manifest_files_from_old_hash_buckets(
    hb_index_copy_by_reference: List[int],
    round_completion_info: RoundCompletionInfo,
    write_to_partition: Partition,
    compacted_manifest: Optional[Manifest] = None,
) -> List[MaterializeResult]:

    manifest_entry_referenced_list = []
    materialize_result_list = []
    hb_index_to_indices = round_completion_info.hb_index_to_entry_range

    if hb_index_to_indices is None:
        logger.info("Nothing to copy by reference. Skipping...")
        return []

    for hb_index in hb_index_copy_by_reference:
        if str(hb_index) not in hb_index_to_indices:
            continue

        indices = hb_index_to_indices[str(hb_index)]
        for offset in range(indices[1] - indices[0]):
            entry_index = indices[0] + offset
            assert entry_index < len(
                compacted_manifest.entries
            ), f"entry index: {entry_index} >= {len(compacted_manifest.entries)}"
            manifest_entry = compacted_manifest.entries[entry_index]
            manifest_entry_referenced_list.append(manifest_entry)

        compacted_manifest = Manifest.of(
            entries=manifest_entry_referenced_list, uuid=str(uuid4())
        )
        delta = Delta.of(
            locator=DeltaLocator.of(write_to_partition.locator),
            delta_type=DeltaType.APPEND,  # Compaction always produces APPEND deltas
            meta=compacted_manifest.meta,
            manifest=compacted_manifest,
            previous_stream_position=write_to_partition.stream_position,
            properties={},
        )
        referenced_pyarrow_write_result = PyArrowWriteResult.of(
            len(manifest_entry_referenced_list),
            compacted_manifest.meta.source_content_length,
            compacted_manifest.meta.content_length,
            compacted_manifest.meta.record_count,
        )
        materialize_result = MaterializeResult.of(
            delta=delta,
            task_index=hb_index,
            pyarrow_write_result=referenced_pyarrow_write_result,
            referenced_pyarrow_write_result=referenced_pyarrow_write_result,
        )
        materialize_result_list.append(materialize_result)
    return materialize_result_list


def _has_previous_compacted_table(input: MergeInput, hb_idx: int) -> bool:
    """
    Checks if the given hash bucket index has a compacted table available from the previous compaction round.

    Args:
        input (MergeInput): The input for the merge operation.
        hb_idx (int): The hash bucket index to check.

    Returns:
        bool: True if the hash bucket index has a compacted table available, False otherwise.
    """
    return (
        input.round_completion_info
        and input.compacted_manifest is not None
        and input.round_completion_info.hb_index_to_entry_range
        and input.round_completion_info.hb_index_to_entry_range.get(str(hb_idx))
        is not None
    )


def _can_copy_by_reference(
    has_delete: bool, merge_file_group: MergeFileGroup, input: MergeInput
) -> bool:
    """
    Can copy by reference only if there are no deletes to merge in
    and previous compacted stream id matches that of new stream
    """
    copy_by_ref = (
        not has_delete
        and not merge_file_group.dfe_groups
        and input.round_completion_info is not None
        and input.compacted_manifest is not None
    )

    if input.disable_copy_by_reference:
        copy_by_ref = False

    logger.info(f"Copy by reference is {copy_by_ref} for {merge_file_group.hb_index}")

    return copy_by_ref


def _flatten_dfe_list(
    df_envelopes_list: List[List[DeltaFileEnvelope]],
) -> List[DeltaFileEnvelope]:
    """
    Flattens a list of lists of DeltaFileEnvelope objects into a single list of DeltaFileEnvelope objects.

    Args:
        df_envelopes_list (List[List[DeltaFileEnvelope]]): A list of lists of DeltaFileEnvelope objects.

    Returns:
        List[DeltaFileEnvelope]: A flattened list of DeltaFileEnvelope objects.
    """
    if not df_envelopes_list:
        return []
    return [d for dfe_list in df_envelopes_list for d in dfe_list]


def _sort_df_envelopes(
    df_envelopes: List[DeltaFileEnvelope],
    key: Callable = lambda df: (df.stream_position, df.file_index),
) -> List[DeltaFileEnvelope]:
    """
    Sorts a list of DeltaFileEnvelope objects based on a specified key function.

    Args:
        df_envelopes (List[DeltaFileEnvelope]): A list of DeltaFileEnvelope objects.
        key (Callable, optional): A function that takes a DeltaFileEnvelope object and returns a key for sorting.
            Defaults to lambda df: (df.stream_position, df.file_index).

    Returns:
        List[DeltaFileEnvelope]: A sorted list of DeltaFileEnvelope objects.
    """
    if not df_envelopes:
        return []
    return sorted(
        df_envelopes,
        key=key,
        reverse=False,  # ascending
    )


def _group_sequence_by_delta_type(
    df_envelopes: List[DeltaFileEnvelope],
) -> Iterator[Tuple[List, List]]:
    """
    Groups a list of DeltaFileEnvelope objects by their delta_type.

    Args:
        df_envelopes (List[DeltaFileEnvelope]): A list of DeltaFileEnvelope objects.

    Yields:
        Iterator[Tuple[DeltaType, List[DeltaFileEnvelope]]]: A tuple containing the delta_type
        and a list of DeltaFileEnvelope objects that share the same delta_type.
    """
    iter_df_envelopes = iter(df_envelopes)
    for delta_type, delta_type_sequence in itertools.groupby(
        iter_df_envelopes, lambda x: x.delta_type
    ):
        yield delta_type, list(delta_type_sequence)


def _compact_tables(
    input: MergeInput,
    dfe_list: Optional[List[List[DeltaFileEnvelope]]],
    hb_idx: int,
    compacted_table: Optional[pa.Table] = None,
) -> Tuple[pa.Table, int, int, int]:
    """
    Compacts a list of DeltaFileEnvelope objects into a single PyArrow table.

    Args:
        input (MergeInput): The input for the merge operation.
        dfe_list (List[List[DeltaFileEnvelope]]): A list of lists of DeltaFileEnvelope objects.
        hb_idx (int): The hash bucket index for the compaction.

    Returns:
        Tuple[pa.Table, int, int, int]: A tuple containing:
            1. The compacted PyArrow table.
            2. The total number of records in the incremental data.
            3. The total number of deduplicated records.
            4. The total number of deleted records due to DELETE operations.
    """
    df_envelopes: List[DeltaFileEnvelope] = _flatten_dfe_list(dfe_list)
    delete_file_envelopes = input.delete_file_envelopes or []
    reordered_all_dfes: List[DeltaFileEnvelope] = _sort_df_envelopes(
        delete_file_envelopes + df_envelopes
    )
    assert all(
        dfe.delta_type in (DeltaType.APPEND, DeltaType.UPSERT, DeltaType.DELETE)
        for dfe in reordered_all_dfes
    ), "All reordered delta file envelopes must be of the APPEND, UPSERT or DELETE"
    table = compacted_table
    aggregated_incremental_len = 0
    aggregated_deduped_records = 0
    aggregated_dropped_records = 0
    for i, (delta_type, delta_type_sequence) in enumerate(
        _group_sequence_by_delta_type(reordered_all_dfes)
    ):
        if delta_type is DeltaType.UPSERT or delta_type is DeltaType.APPEND:
            (table, incremental_len, deduped_records, merge_time,) = _apply_upserts(
                input=input,
                dfe_list=delta_type_sequence,
                hb_idx=hb_idx,
                prev_table=table,
            )
            logger.info(
                f" [Merge task index {input.merge_task_index}] Merged"
                f" record count: {len(table)}, size={table.nbytes} took: {merge_time}s"
            )
            aggregated_incremental_len += incremental_len
            aggregated_deduped_records += deduped_records
        elif delta_type is DeltaType.DELETE:
            table_size_before_delete = len(table) if table else 0
            (table, dropped_rows), delete_time = timed_invocation(
                func=input.delete_strategy.apply_many_deletes,
                table=table,
                delete_file_envelopes=delta_type_sequence,
            )
            logger.info(
                f" [Merge task index {input.merge_task_index}]"
                + f" Dropped record count: {dropped_rows} from table"
                + f" of record count {table_size_before_delete} took: {delete_time}s"
            )
            aggregated_dropped_records += dropped_rows
    return (
        table,
        aggregated_incremental_len,
        aggregated_deduped_records,
        aggregated_dropped_records,
    )


def _apply_upserts(
    input: MergeInput,
    dfe_list: List[DeltaFileEnvelope],
    hb_idx,
    prev_table=None,
) -> Tuple[pa.Table, int, int, int]:
    assert all(
        dfe.delta_type is DeltaType.UPSERT or dfe.delta_type is DeltaType.APPEND
        for dfe in dfe_list
    ), "All incoming delta file envelopes must of the DeltaType.UPSERT or DeltaType.APPEND"
    logger.info(
        f"[Hash bucket index {hb_idx}] Reading dedupe input for "
        f"{len(dfe_list)} delta file envelope lists..."
    )
    table = _build_incremental_table(dfe_list)
    incremental_len = len(table)
    logger.info(
        f"[Hash bucket index {hb_idx}] Got the incremental table of length {incremental_len}"
    )
    if input.sort_keys:
        # Incremental is sorted and merged, as sorting
        # on non event based sort key does not produce consistent
        # compaction results. E.g., compaction(delta1, delta2, delta3)
        # will not be equal to compaction(compaction(delta1, delta2), delta3).
        table = table.sort_by(
            [pa_key for key in input.sort_keys for pa_key in key.arrow]
        )
    hb_table_record_count = len(table) + (len(prev_table) if prev_table else 0)
    table, merge_time = timed_invocation(
        func=_merge_tables,
        table=table,
        primary_keys=input.primary_keys,
        can_drop_duplicates=input.drop_duplicates,
        hb_index=hb_idx,
        num_buckets=input.hash_bucket_count,
        original_fields=input.original_fields,
        compacted_table=prev_table,
    )
    deduped_records = hb_table_record_count - len(table)
    return table, incremental_len, deduped_records, merge_time


def _copy_manifests_from_hash_bucketing(
    input: MergeInput, hb_index_copy_by_reference_ids: List[int]
) -> List[MaterializeResult]:
    materialized_results: List[MaterializeResult] = []

    if input.round_completion_info:
        referenced_materialized_results = (
            _copy_all_manifest_files_from_old_hash_buckets(
                hb_index_copy_by_reference_ids,
                input.round_completion_info,
                input.write_to_partition,
                input.compacted_manifest,
            )
        )
        logger.info(
            f"Copying {len(referenced_materialized_results)} manifest files by reference..."
        )
        materialized_results.extend(referenced_materialized_results)

    return materialized_results


@success_metric(name=MERGE_SUCCESS_COUNT)
@failure_metric(name=MERGE_FAILURE_COUNT)
@categorize_errors
def _timed_merge(input: MergeInput) -> MergeResult:
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with (
        memray.Tracker(f"merge_{worker_id}_{task_id}.bin")
        if input.enable_profiler
        else nullcontext()
    ):
        total_input_records, total_deduped_records = 0, 0
        total_dropped_records = 0
        materialized_results: List[MaterializeResult] = []
        merge_file_groups = input.merge_file_groups_provider.create()
        hb_index_copy_by_ref_ids = []

        for merge_file_group in merge_file_groups:
            compacted_table = None
            has_delete = input.delete_file_envelopes is not None
            if has_delete:
                assert (
                    input.delete_strategy is not None
                ), "Merge input missing delete_strategy"
            if _can_copy_by_reference(
                has_delete=has_delete, merge_file_group=merge_file_group, input=input
            ):
                hb_index_copy_by_ref_ids.append(merge_file_group.hb_index)
                continue
            if _has_previous_compacted_table(input, merge_file_group.hb_index):
                compacted_table = _download_compacted_table(
                    hb_index=merge_file_group.hb_index,
                    rci=input.round_completion_info,
                    primary_keys=input.primary_keys,
                    all_column_names=input.all_column_names,
                    compacted_delta_manifest=input.compacted_manifest,
                    read_kwargs_provider=input.read_kwargs_provider,
                    deltacat_storage=input.deltacat_storage,
                    deltacat_storage_kwargs=input.deltacat_storage_kwargs,
                )
            if not merge_file_group.dfe_groups and compacted_table is None:
                logger.warning(
                    f" [Hash bucket index {merge_file_group.hb_index}]"
                    + f" No new deltas and no compacted table found. Skipping compaction for {merge_file_group.hb_index}"
                )
                continue
            table, input_records, deduped_records, dropped_records = _compact_tables(
                input,
                merge_file_group.dfe_groups,
                merge_file_group.hb_index,
                compacted_table,
            )
            total_input_records += input_records
            total_deduped_records += deduped_records
            total_dropped_records += dropped_records
            materialized_results.append(
                merge_utils.materialize(input, merge_file_group.hb_index, [table])
            )

        if hb_index_copy_by_ref_ids:
            materialized_results.extend(
                _copy_manifests_from_hash_bucketing(input, hb_index_copy_by_ref_ids)
            )

        logger.info(
            f"[Hash group index: {input.merge_file_groups_provider.hash_group_index}]"
            f" Total number of materialized results produced: {len(materialized_results)} "
        )

        peak_memory_usage_bytes = get_current_process_peak_memory_usage_in_bytes()
        logger.info(
            f"Peak memory usage in bytes after merge: {peak_memory_usage_bytes}"
        )

        return MergeResult(
            materialized_results,
            np.int64(total_input_records),
            np.int64(total_deduped_records),
            np.int64(total_dropped_records),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def merge(input: MergeInput) -> MergeResult:
    with ProcessUtilizationOverTimeRange() as process_util:
        logger.info(f"Starting merge task {input.merge_task_index}...")

        # Log node peak memory utilization every 10 seconds
        def log_peak_memory():
            logger.debug(
                f"Process peak memory utilization so far: {process_util.max_memory} bytes "
                f"({process_util.max_memory/BYTES_PER_GIBIBYTE} GB)"
            )

        if input.memory_logs_enabled:
            process_util.schedule_callback(log_peak_memory, 10)

        merge_result, duration = timed_invocation(func=_timed_merge, input=input)

        emit_metrics_time = 0.0
        if input.metrics_config:
            emit_result, latency = timed_invocation(
                func=emit_timer_metrics,
                metrics_name=MERGE_TIME_IN_SECONDS,
                value=duration,
                metrics_config=input.metrics_config,
            )
            emit_metrics_time = latency

        logger.info(f"Finished merge task {input.merge_task_index}...")
        return MergeResult(
            merge_result[0],
            merge_result[1],
            merge_result[2],
            merge_result[3],
            merge_result[4],
            np.double(emit_metrics_time),
            merge_result[6],
        )



from __future__ import annotations

import botocore.exceptions

from daft.exceptions import DaftTransientError
from deltacat.utils.common import env_string, env_bool
from deltacat.utils.common import env_integer

# Environment variables
DELTACAT_SYS_LOG_LEVEL = env_string("DELTACAT_SYS_LOG_LEVEL", "DEBUG")
DELTACAT_SYS_LOG_DIR = env_string(
    "DELTACAT_SYS_LOG_DIR",
    "/tmp/deltacat/var/output/logs/",
)
DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME",
    "deltacat-python.info.log",
)
DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME",
    "deltacat-python.debug.log",
)

DELTACAT_APP_LOG_LEVEL = env_string("DELTACAT_APP_LOG_LEVEL", "DEBUG")
DELTACAT_APP_LOG_DIR = env_string(
    "DELTACAT_APP_LOG_DIR",
    "/tmp/deltacat/var/output/logs/",
)
DELTACAT_APP_INFO_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_APP_INFO_LOG_BASE_FILE_NAME",
    "application.info.log",
)
DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME = env_string(
    "DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME",
    "application.debug.log",
)
# A json context which will be logged along with other context args.
DELTACAT_LOGGER_CONTEXT = env_string("DELTACAT_LOGGER_CONTEXT", None)
DELTACAT_LOGGER_USE_SINGLE_HANDLER = env_bool(
    "DELTACAT_LOGGER_USE_SINGLE_HANDLER",
    False,
)
DELTACAT_ROOT = env_string(
    "DELTACAT_ROOT",
    "",
)

# CLI Args
METAFILE_FORMAT_JSON = "json"
METAFILE_FORMAT_MSGPACK = "msgpack"
METAFILE_FORMAT = env_string("METAFILE_FORMAT", METAFILE_FORMAT_MSGPACK)
SUPPORTED_METAFILE_FORMATS = [METAFILE_FORMAT_JSON, METAFILE_FORMAT_MSGPACK]
METAFILE_EXT = {
    "json": ".json",
    "msgpack": ".mpk",
}[METAFILE_FORMAT]


# Byte Units
BYTES_PER_KIBIBYTE = 2**10
BYTES_PER_MEBIBYTE = 2**20
BYTES_PER_GIBIBYTE = 2**30
BYTES_PER_TEBIBYTE = 2**40
BYTES_PER_PEBIBYTE = 2**50

SIGNED_INT64_MIN_VALUE = -(2**63)
SIGNED_INT64_MAX_VALUE = 2**63 - 1

# Time Units
NANOS_PER_SEC = 1_000_000_000
MICROS_PER_SEC = 1_000_000
MILLIS_PER_SEC = 1000

# Inflation multiplier from snappy-compressed parquet to pyarrow.
# This should be kept larger than actual average inflation multipliers.
# Note that this is a very rough guess since actual observed pyarrow
# inflation multiplier for snappy-compressed parquet is about 5.45X for
# all rows, but here we're trying to guess the inflation multipler for just
# a primary key SHA1 digest and sort key columns (which could be all columns
# of the table in the worst case, but here we're assuming that they
# represent no more than ~1/4th of the total table bytes)
PYARROW_INFLATION_MULTIPLIER = 2.5

# Inflation multiplier from snappy-compressed parquet to pyarrow for all columns.
PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS = 6

MEMORY_TO_HASH_BUCKET_COUNT_RATIO = 0.0512 * BYTES_PER_TEBIBYTE

# The number of bytes allocated to null values in string physical type in parquet
NULL_SIZE_BYTES = 4

# Metastore Constants
REVISION_DIR_NAME: str = "rev"
TXN_DIR_NAME: str = "txn"
RUNNING_TXN_DIR_NAME: str = "running"
FAILED_TXN_DIR_NAME: str = "failed"
PAUSED_TXN_DIR_NAME: str = "paused"
SUCCESS_TXN_DIR_NAME: str = "success"
DATA_FILE_DIR_NAME: str = "data"
REV_DIR_NAME: str = "rev"
TXN_PART_SEPARATOR = "_"

# Storage interface defaults
# These defaults should be applied in catalog interface implementations
# Storage interface implementations should be agnostic to defaults and require full information
DEFAULT_CATALOG = "default"
DEFAULT_NAMESPACE = "default"
DEFAULT_TABLE_VERSION = "1"
DEFAULT_STREAM_ID = "stream"
DEFAULT_PARTITION_ID = "partition"
DEFAULT_PARTITION_VALUES = ["default"]

# Transaction Status constants
SUCCESSFULLY_CLEANED = "cleaned"
CURRENTLY_CLEANING = "cleaning"
TIMEOUT_TXN = "timedout"

# operation timeout constants
OPERATION_TIMEOUTS = {
    "create": 5,
    "update": 3,
    "delete": 4,
    "read_siblings": 2,
    "read_children": 2,
    "read_latest": 3,
    "read_exists": 1,
}
# Upload/Download Retry Defaults
UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY = env_integer(
    "UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY", 10 * 60
)
UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY = env_integer(
    "UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY", 30 * 60
)
DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY = env_integer(
    "DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY", 30 * 60
)
DEFAULT_FILE_READ_TIMEOUT_MS = env_integer(
    "DEFAULT_FILE_READ_TIMEOUT_MS", 300_000
)  # 5 mins
RETRYABLE_TRANSIENT_ERRORS = (
    OSError,
    botocore.exceptions.ConnectionError,
    botocore.exceptions.HTTPClientError,
    botocore.exceptions.NoCredentialsError,
    botocore.exceptions.ConnectTimeoutError,
    botocore.exceptions.ReadTimeoutError,
    DaftTransientError,
)



from deltacat.storage import (
    Delta,
)

from typing import List

from dataclasses import dataclass, fields
from deltacat.compute.compactor_v2.deletes.delete_file_envelope import (
    DeleteFileEnvelope,
)
from deltacat.compute.compactor_v2.deletes.delete_strategy import (
    DeleteStrategy,
)


@dataclass
class PrepareDeleteResult:
    non_delete_deltas: List[Delta]
    delete_file_envelopes: List[DeleteFileEnvelope]
    delete_strategy: DeleteStrategy

    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))



import pytest

from deltacat.constants import BYTES_PER_KIBIBYTE
from deltacat.storage.model.table_version import TableVersion, TableVersionLocator


@pytest.mark.parametrize(
    "previous_version, expected_next_version",
    [
        (None, "1"),
        ("v.1", "v.2"),
        ("1", "2"),
        ("1.0", "1.1"),
        ("v.999", "v.1000"),
    ],
)
def test_next_version(previous_version, expected_next_version):
    new_version = TableVersion.next_version(previous_version)
    assert isinstance(new_version, str)
    assert new_version == expected_next_version


@pytest.mark.parametrize(
    "table_version, expected_parsed_version",
    [
        ("v.1", ("v.", 1)),
        ("1", (None, 1)),
        ("1.0", ("1.", 0)),
        ("v.999", ("v.", 999)),
    ],
)
def test_parse_version(table_version, expected_parsed_version):
    prefix, version_number = TableVersion.parse_table_version(table_version)
    if prefix is not None:
        assert isinstance(prefix, str)
    assert isinstance(version_number, int)
    assert (prefix, version_number) == expected_parsed_version


def test_version_validation_invalid():
    with pytest.raises(ValueError):
        TableVersionLocator.at(
            namespace="test_namespace",
            table_name="test_table",
            table_version="invalid_version",
        )


def test_version_validation_valid_to_invalid():
    valid_tv_locator = TableVersionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="1",
    )
    assert valid_tv_locator.table_version == "1"
    tv = TableVersion.of(
        locator=valid_tv_locator,
        schema=None,
    )
    assert tv.current_version_number() == 1
    with pytest.raises(ValueError):
        valid_tv_locator.table_version = "invalid_version"


def test_version_validation_numeric_name():
    valid_tv_locator = TableVersionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="1.0",
    )
    assert valid_tv_locator.table_version == "1.0"
    tv = TableVersion.of(
        locator=valid_tv_locator,
        schema=None,
    )
    assert tv.current_version_number() == 0


def test_version_validation_truncate_leading_zeros():
    valid_tv_locator = TableVersionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version="1.00002",
    )
    # ensure that leading 0's are truncated
    assert valid_tv_locator.table_version == "1.2"
    tv = TableVersion.of(
        locator=valid_tv_locator,
        schema=None,
    )
    assert tv.current_version_number() == 2


def test_version_validation_version_id_length_limits():
    # ensure that long version identifiers are accepted
    long_tv_id = "a" * (BYTES_PER_KIBIBYTE - 2) + ".1"
    valid_tv_locator = TableVersionLocator.at(
        namespace="test_namespace",
        table_name="test_table",
        table_version=long_tv_id,
    )
    assert valid_tv_locator.table_version == long_tv_id
    tv = TableVersion.of(
        locator=valid_tv_locator,
        schema=None,
    )
    assert tv.current_version_number() == 1
    # ensure that an excessively long version identifier is rejected
    with pytest.raises(ValueError):
        valid_tv_locator.table_version = "0" * (BYTES_PER_KIBIBYTE + 1)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import List, TypeVar, Any

import pyarrow as pa

from deltacat.storage.model.schema import FieldName


R = TypeVar("R")
C = TypeVar("C")


@dataclass(frozen=True)
class Expression(ABC):
    """
    This abstract class serves as the foundation for all expression types
    in the Deltacat expression tree. It defines common functionality and structure
    for expressions.
    """

    def __str__(self) -> str:
        """Convert the expression to a string representation.

        Returns:
            str: A string representation of the expression.
        """
        from deltacat.storage.model.expression.visitor import DisplayVisitor

        return DisplayVisitor().visit(self, None)


@dataclass(frozen=True)
class UnaryExpression(Expression, ABC):
    """
    This abstract class represents operations that work on a single input
    expression (e.g., NOT, IS NULL).

    Args:
        operand: The expression this unary operation applies to.
    """

    operand: Expression

    @classmethod
    def of(cls, operand: Expression | Any) -> UnaryExpression:
        """
        Create a new instance with the given operand.

        Args:
            operand: The expression or value to use as the operand.
                If not an Expression, it will be converted to a Literal.

        Returns:
            UnaryExpression: A new instance of this unary expression.
        """
        operand_expr = (
            operand if isinstance(operand, Expression) else Literal(pa.scalar(operand))
        )
        return cls(operand_expr)


@dataclass(frozen=True)
class BinaryExpression(Expression, ABC):
    """
    This abstract class represents operations that work on two input
    expressions (e.g., AND, OR, comparison operators).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    left: Expression
    right: Expression

    def with_left(self, left: Expression) -> BinaryExpression:
        """
        Create a new expression with a different left operand.

        Args:
            left: The new left operand expression.

        Returns:
            BinaryExpression: A new binary expression with the updated left operand.
        """
        return type(self)(left, self.right)

    def with_right(self, right: Expression) -> BinaryExpression:
        """
        Create a new expression with a different right operand.

        Args:
            right: The new right operand expression.

        Returns:
            BinaryExpression: A new binary expression with the updated right operand.
        """
        return type(self)(self.left, right)

    @classmethod
    def of(cls, left: Expression | Any, right: Expression | Any) -> BinaryExpression:
        """
        Create a new instance with the given operands.

        Args:
            left: The left operand expression or value.
                If not an Expression, it will be converted to a Literal.
            right: The right operand expression or value.
                If not an Expression, it will be converted to a Literal.

        Returns:
            BinaryExpression: A new instance of this binary expression.
        """
        left_expr = left if isinstance(left, Expression) else Literal(pa.scalar(left))
        right_expr = (
            right if isinstance(right, Expression) else Literal(pa.scalar(right))
        )
        return cls(left_expr, right_expr)


@dataclass(frozen=True)
class BooleanExpression(Expression, ABC):
    """
    This abstract class represents expressions that produce a boolean
    value when evaluated, such as comparisons, logical operations, etc.
    """

    def and_(self, other: BooleanExpression | Any) -> BooleanExpression:
        """
        Combine this expression with another using logical AND.

        Args:
            other: The expression to AND with this one.
                If not a BooleanExpression, it will be converted to a Literal.

        Returns:
            BooleanExpression: A new AND expression.
        """
        other_expr = (
            other if isinstance(other, BooleanExpression) else Literal(pa.scalar(other))
        )
        return And(self, other_expr)

    def or_(self, other: BooleanExpression | Any) -> BooleanExpression:
        """
        Combine this expression with another using logical OR.

        Args:
            other: The expression to OR with this one.
                If not a BooleanExpression, it will be converted to a Literal.

        Returns:
            BooleanExpression: A new OR expression.
        """
        other_expr = (
            other if isinstance(other, BooleanExpression) else Literal(pa.scalar(other))
        )
        return Or(self, other_expr)

    def not_(self) -> BooleanExpression:
        """
        Negate this expression with logical NOT.

        Returns:
            BooleanExpression: A new NOT expression.
        """
        return Not(self)


@dataclass(frozen=True)
class Reference(Expression):
    """
    This class represents a reference to a column or field in a dataset.

    TODO: Add validation and deltacat schema binding.

    Args:
        field: The field name of the referenced field.
        index: Optional index value for the field in the schema.
    """

    field: FieldName
    index: int | None = None

    @classmethod
    def of(cls, field: FieldName) -> Reference:
        """
        Create a new reference.

        Args:
            field: The field name of a Deltacat field.

        Returns:
            Reference: A new reference expression.
        """
        return cls(field)

    def eq(self, other: Expression | Any) -> Equal:
        """
        Create an equality comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Equal: A new equality comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return Equal(self, other)

    def ne(self, other: Expression | Any) -> NotEqual:
        """
        Create an inequality comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            NotEqual: A new inequality comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return NotEqual(self, other)

    def gt(self, other: Expression | Any) -> GreaterThan:
        """
        Create a greater than comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            GreaterThan: A new greater than comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return GreaterThan(self, other)

    def lt(self, other: Expression | Any) -> LessThan:
        """
        Create a less than comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            LessThan: A new less than comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return LessThan(self, other)

    def ge(self, other: Expression | Any) -> GreaterThanEqual:
        """
        Create a greater than or equal comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            GreaterThanEqual: A new greater than or equal comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return GreaterThanEqual(self, other)

    def le(self, other: Expression | Any) -> LessThanEqual:
        """
        Create a less than or equal comparison with this reference.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            LessThanEqual: A new less than or equal comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return LessThanEqual(self, other)

    def is_null(self) -> IsNull:
        """
        Create an IS NULL check for this reference.

        Returns:
            IsNull: A new IS NULL expression.
        """
        return IsNull(self)

    def in_(self, values: List[Expression | Any]) -> In:
        """
        Create an IN check for this reference.

        Args:
            values: List of expressions or values to check against.
                Non-Expression values will be converted to Literals.

        Returns:
            In: A new IN expression.
        """
        expr_values = [
            v if isinstance(v, Expression) else Literal(pa.scalar(v)) for v in values
        ]
        return In(self, expr_values)

    def between(self, lower: Expression | Any, upper: Expression | Any) -> Between:
        """
        Create a BETWEEN check for this reference.

        Args:
            lower: The lower bound expression or value.
                If not an Expression, it will be converted to a Literal.
            upper: The upper bound expression or value.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Between: A new BETWEEN expression.
        """
        lower_expr = (
            lower if isinstance(lower, Expression) else Literal(pa.scalar(lower))
        )
        upper_expr = (
            upper if isinstance(upper, Expression) else Literal(pa.scalar(upper))
        )
        return Between(self, lower_expr, upper_expr)

    def like(self, pattern: str | Expression) -> Like:
        """
        Create a LIKE pattern match for this reference.

        Args:
            pattern: The pattern to match against.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Like: A new LIKE expression.
        """
        pattern_expr = (
            pattern if isinstance(pattern, Expression) else Literal(pa.scalar(pattern))
        )
        return Like(self, pattern_expr)


@dataclass(frozen=True)
class Literal(Expression):
    """
    Literal value using PyArrow scalar.

    TODO: implement custom deltacat types and literals.

    Args:
        value: The PyArrow scalar value.
    """

    value: pa.Scalar

    @classmethod
    def of(cls, value: Any) -> Literal:
        """
        Create a new literal value.

        Args:
            value: The value to create a literal for.
                If not a PyArrow Scalar, it will be converted.

        Returns:
            Literal: A new literal expression.
        """
        if not isinstance(value, pa.Scalar):
            value = pa.scalar(value)
        return cls(value)

    def eq(self, other: Expression | Any) -> Equal:
        """
        Create an equality comparison with this literal.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Equal: A new equality comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return Equal(self, other)

    def ne(self, other: Expression | Any) -> NotEqual:
        """
        Create an inequality comparison with this literal.

        Args:
            other: The expression or value to compare with.
                If not an Expression, it will be converted to a Literal.

        Returns:
            NotEqual: A new inequality comparison expression.
        """
        if not isinstance(other, Expression):
            other = Literal(pa.scalar(other))
        return NotEqual(self, other)


@dataclass(frozen=True)
class Equal(BinaryExpression, BooleanExpression):
    """
    Equality comparison (=).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class NotEqual(BinaryExpression, BooleanExpression):
    """
    Inequality comparison (<>).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class GreaterThan(BinaryExpression, BooleanExpression):
    """
    Greater than comparison (>).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class LessThan(BinaryExpression, BooleanExpression):
    """
    Less than comparison (<).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class GreaterThanEqual(BinaryExpression, BooleanExpression):
    """
    Greater than or equal comparison (>=).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class LessThanEqual(BinaryExpression, BooleanExpression):
    """
    Less than or equal comparison (<=).

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class And(BinaryExpression, BooleanExpression):
    """
    Logical AND operation.

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class Or(BinaryExpression, BooleanExpression):
    """
    Logical OR operation.

    Args:
        left: The left operand expression.
        right: The right operand expression.
    """

    pass


@dataclass(frozen=True)
class Not(UnaryExpression, BooleanExpression):
    """
    Logical NOT operation.

    Args:
        operand: The boolean expression to negate.
    """

    operand: BooleanExpression


@dataclass(frozen=True)
class In(BooleanExpression):
    """
    Checks if a value is in a list of values.

    Args:
        value: The expression to check.
        values: List of expressions to check against.
    """

    value: Expression
    values: List[Expression] = field(default_factory=list)

    @classmethod
    def of(cls, value: Expression | Any, values: List[Expression | Any]) -> In:
        """
        Create a new IN expression.

        Args:
            value: The expression or value to check.
                If not an Expression, it will be converted to a Literal.
            values: List of expressions or values to check against.
                Non-Expression values will be converted to Literals.

        Returns:
            In: A new IN expression.
        """
        value_expr = (
            value if isinstance(value, Expression) else Literal(pa.scalar(value))
        )
        value_exprs = [
            v if isinstance(v, Expression) else Literal(pa.scalar(v)) for v in values
        ]
        return cls(value_expr, value_exprs)


@dataclass(frozen=True)
class Between(BooleanExpression):
    """
    Checks if a value is between two other values (inclusive).

    Args:
        value: The expression to check.
        lower: The lower bound expression.
        upper: The upper bound expression.
    """

    value: Expression
    lower: Expression
    upper: Expression

    @classmethod
    def of(
        cls, value: Expression | Any, lower: Expression | Any, upper: Expression | Any
    ) -> Between:
        """
        Create a new BETWEEN expression.

        Args:
            value: The expression or value to check.
                If not an Expression, it will be converted to a Literal.
            lower: The lower bound expression or value.
                If not an Expression, it will be converted to a Literal.
            upper: The upper bound expression or value.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Between: A new BETWEEN expression.
        """
        value_expr = (
            value if isinstance(value, Expression) else Literal(pa.scalar(value))
        )
        lower_expr = (
            lower if isinstance(lower, Expression) else Literal(pa.scalar(lower))
        )
        upper_expr = (
            upper if isinstance(upper, Expression) else Literal(pa.scalar(upper))
        )
        return cls(value_expr, lower_expr, upper_expr)


@dataclass(frozen=True)
class Like(BooleanExpression):
    """
    Pattern matching for string values.

    Args:
        value: The expression to check.
        pattern: The pattern to match against.
    """

    value: Expression
    pattern: Expression

    @classmethod
    def of(cls, value: Expression | Any, pattern: Expression | Any) -> Like:
        """
        Create a new LIKE expression.

        Args:
            value: The expression or value to check.
                If not an Expression, it will be converted to a Literal.
            pattern: The pattern expression or value to match against.
                If not an Expression, it will be converted to a Literal.

        Returns:
            Like: A new LIKE expression.
        """
        value_expr = (
            value if isinstance(value, Expression) else Literal(pa.scalar(value))
        )
        pattern_expr = (
            pattern if isinstance(pattern, Expression) else Literal(pa.scalar(pattern))
        )
        return cls(value_expr, pattern_expr)


@dataclass(frozen=True)
class IsNull(UnaryExpression, BooleanExpression):
    """
    Checks if a value is NULL.

    Args:
        operand: The expression to check for NULL.
    """

    pass



import time
from typing import List, Optional, Tuple

from deltacat.compute.compactor import (
    DeltaAnnotated,
    DeltaFileEnvelope,
)
from deltacat.storage import (
    Delta,
)
from deltacat.storage.model.delta import DeltaType
from deltacat.storage import metastore
from deltacat.types.media import StorageType
from deltacat.utils.common import ReadKwargsProvider
from deltacat import logs

import pyarrow as pa
import logging


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def contains_delete_deltas(deltas: List[Delta]) -> bool:
    for delta in deltas:
        if delta.type is DeltaType.DELETE:
            return True
    return False


def read_delta_file_envelopes(
    annotated_delta: DeltaAnnotated,
    all_column_names: List[str],
    read_kwargs_provider: Optional[ReadKwargsProvider],
    deltacat_storage: metastore = metastore,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> Tuple[Optional[List[DeltaFileEnvelope]], int, int]:
    tables = deltacat_storage.download_delta(
        annotated_delta,
        max_parallelism=1,
        file_reader_kwargs_provider=read_kwargs_provider,
        storage_type=StorageType.LOCAL,
        all_column_names=all_column_names,
        **deltacat_storage_kwargs,
    )
    annotations = annotated_delta.annotations
    assert len(tables) == len(annotations), (
        f"Unexpected Error: Length of downloaded delta manifest tables "
        f"({len(tables)}) doesn't match the length of delta manifest "
        f"annotations ({len(annotations)})."
    )
    if not tables:
        return None, 0, 0

    delta_stream_position = annotations[0].annotation_stream_position
    delta_type = annotations[0].annotation_delta_type

    for annotation in annotations:
        assert annotation.annotation_stream_position == delta_stream_position, (
            f"Annotation stream position does not match - {annotation.annotation_stream_position} "
            f"!= {delta_stream_position}"
        )
        assert annotation.annotation_delta_type == delta_type, (
            f"Annotation delta type does not match - {annotation.annotation_delta_type} "
            f"!= {delta_type}"
        )

    delta_file_envelopes = []
    table = pa.concat_tables(tables)
    total_record_count = len(table)
    total_size_bytes = int(table.nbytes)

    delta_file = DeltaFileEnvelope.of(
        stream_position=delta_stream_position,
        delta_type=delta_type,
        table=table,
    )
    delta_file_envelopes.append(delta_file)
    return delta_file_envelopes, total_record_count, total_size_bytes


def get_local_delta_file_envelopes(
    uniform_deltas: List[DeltaAnnotated],
    read_kwargs_provider: Optional[ReadKwargsProvider],
    deltacat_storage=metastore,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> Tuple[List[DeltaFileEnvelope], int]:
    local_dfe_list = []
    input_records_count = 0
    logger.info(f"Getting {len(uniform_deltas)} DFE Tasks.")
    dfe_start = time.monotonic()
    for annotated_delta in uniform_deltas:
        (
            delta_file_envelopes,
            total_record_count,
            total_size_bytes,
        ) = read_delta_file_envelopes(
            annotated_delta,
            read_kwargs_provider,
            deltacat_storage,
            deltacat_storage_kwargs,
        )
        if delta_file_envelopes:
            local_dfe_list.extend(delta_file_envelopes)
            input_records_count += total_record_count
    dfe_end = time.monotonic()
    logger.info(f"Retrieved {len(local_dfe_list)} DFE Tasks in {dfe_end - dfe_start}s.")
    return local_dfe_list, input_records_count



import argparse
import os
import logging
from typing import Dict, Any

from deltacat import logs

from deltacat.constants import (
    DELTACAT_APP_LOG_LEVEL,
    DELTACAT_SYS_LOG_LEVEL,
    DELTACAT_APP_LOG_DIR,
    DELTACAT_SYS_LOG_DIR,
    DELTACAT_APP_INFO_LOG_BASE_FILE_NAME,
    DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME,
    DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME,
    DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME,
    DELTACAT_LOGGER_USE_SINGLE_HANDLER,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def create_ray_runtime_environment() -> Dict[str, Any]:
    # log the system environment for debugging
    logger.debug(f"System Environment: {os.environ}")

    # read the stage (e.g. alpha, beta, dev, etc.) from system environment vars
    stage = os.environ.get("STAGE")
    logger.debug(f"Runtime Environment Stage: {stage}")
    runtime_environment = None
    if stage:
        worker_env_vars = {
            # forward the STAGE environment variable to workers
            "STAGE": stage,
            # forward deltacat logging environment variables to workers
            "DELTACAT_APP_LOG_LEVEL": DELTACAT_APP_LOG_LEVEL,
            "DELTACAT_SYS_LOG_LEVEL": DELTACAT_SYS_LOG_LEVEL,
            "DELTACAT_APP_LOG_DIR": DELTACAT_APP_LOG_DIR,
            "DELTACAT_SYS_LOG_DIR": DELTACAT_SYS_LOG_DIR,
            "DELTACAT_APP_INFO_LOG_BASE_FILE_NAME": DELTACAT_APP_INFO_LOG_BASE_FILE_NAME,
            "DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME": DELTACAT_SYS_INFO_LOG_BASE_FILE_NAME,
            "DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME": DELTACAT_APP_DEBUG_LOG_BASE_FILE_NAME,
            "DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME": DELTACAT_SYS_DEBUG_LOG_BASE_FILE_NAME,
            "DELTACAT_LOGGER_USE_SINGLE_HANDLER": str(
                DELTACAT_LOGGER_USE_SINGLE_HANDLER
            ),
        }
        # setup runtime environment from system environment variables:
        runtime_environment = {
            "env_vars": worker_env_vars,
        }
    return runtime_environment


def store_cli_args_in_os_environ(script_args_list=[]):
    parser = argparse.ArgumentParser()
    for args, kwargs in script_args_list:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"Command Line Arguments: {args}")
    os.environ.update(vars(args))



import pytest
import pyarrow as pa
from deltacat.exceptions import SchemaValidationError

from deltacat.storage.model.schema import (
    Schema,
    Field,
    BASE_SCHEMA_NAME,
    SchemaConsistencyType,
    SchemaUpdate,
)


@pytest.fixture
def schema_a():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )


@pytest.fixture
def schema_b():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col2", pa.string(), nullable=True),
                field_id=2,
                is_merge_key=False,
            )
        ]
    )


@pytest.fixture
def schema_c():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col3", pa.float64(), nullable=False),
                field_id=3,
                is_merge_key=False,
            )
        ]
    )


@pytest.fixture
def schema_d():
    schema = Schema.of(
        [
            Field.of(
                field=pa.field("col_named", pa.int64(), nullable=True),
                field_id=4,
                is_merge_key=False,
            )
        ]
    )
    return schema


def test_of_with_dict(schema_a, schema_b):
    input_dict = {"schema_a": schema_a.arrow, "schema_b": schema_b.arrow}
    schema = Schema.of(input_dict)

    assert isinstance(schema, Schema)
    assert list(schema.subschemas.keys()) == ["schema_a", "schema_b"]
    assert list(schema.subschemas_to_field_ids.keys()) == ["schema_a", "schema_b"]
    assert schema.subschemas["schema_a"].equivalent_to(schema_a, True)
    assert schema.subschemas["schema_b"].equivalent_to(schema_b, True)


def test_of_invalid_input():
    with pytest.raises(ValueError):
        Schema.of(42)
    with pytest.raises(ValueError):
        Schema.of(["foo"])
    with pytest.raises(ValueError):
        Schema.of({"foo": [42]})


def test_insert_explicit_name(schema_a, schema_b):
    schema = Schema.of(schema_a.arrow)
    assert not schema.subschemas
    new_schema = schema.add_subschema("explicit", schema_b.arrow)
    assert not schema.subschemas
    assert "explicit" in new_schema.subschemas
    assert "explicit" in new_schema.subschemas_to_field_ids
    assert BASE_SCHEMA_NAME in new_schema.subschemas
    assert BASE_SCHEMA_NAME in new_schema.subschemas_to_field_ids
    assert new_schema.subschemas[BASE_SCHEMA_NAME].equivalent_to(schema_a)
    assert new_schema.subschemas["explicit"].equivalent_to(schema_b)


def test_insert_reserved_name_fails(schema_a, schema_b):
    schema = Schema.of(schema_a.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema(BASE_SCHEMA_NAME, schema_b.arrow)


def test_insert_duplicate_schema_name_fails(schema_a, schema_b):
    schema = Schema.of({"dupe": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.add_subschema("dupe", schema_b.arrow)


def test_insert_duplicate_field_name_fails(schema_a):
    schema = Schema.of(schema_a.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_name", schema_a.arrow)


def test_insert_autofill_field_id():
    schema1 = [
        Field.of(
            field=pa.field("col1", pa.int32(), nullable=False),
            is_merge_key=True,
        )
    ]
    schema2 = [
        Field.of(
            field=pa.field("col2", pa.int32(), nullable=False),
            is_merge_key=True,
        )
    ]
    schema = Schema.of(schema1)
    new_schema = schema.add_subschema("explicit", schema2)
    assert not schema.subschemas
    assert "explicit" in new_schema.subschemas
    assert "explicit" in new_schema.subschemas_to_field_ids
    assert BASE_SCHEMA_NAME in new_schema.subschemas
    assert BASE_SCHEMA_NAME in new_schema.subschemas_to_field_ids
    assert len(new_schema.subschemas[BASE_SCHEMA_NAME].fields) == 1
    assert len(new_schema.subschemas["explicit"].fields) == 1
    assert new_schema.subschemas[BASE_SCHEMA_NAME].fields[0].id == 0
    assert new_schema.subschemas["explicit"].fields[0].id == 1
    assert new_schema.subschemas[BASE_SCHEMA_NAME].equivalent_to(Schema.of(schema1))
    schema2_with_expected_field_id = [
        Field.of(
            field=pa.field("col2", pa.int32(), nullable=False),
            field_id=1,
            is_merge_key=True,
        )
    ]
    assert new_schema.subschemas["explicit"].equivalent_to(
        Schema.of(schema2_with_expected_field_id)
    )


def test_insert_duplicate_field_name_case_insensitive_fails():
    schema1 = Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                field=pa.field("COL1", pa.int32(), nullable=False),
                field_id=2,
                is_merge_key=True,
            )
        ]
    )
    schema = Schema.of(schema1.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_name_case_insensitive", schema2.arrow)


def test_insert_duplicate_field_id_fails():
    schema1 = Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                field=pa.field("col2", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema = Schema.of(schema1.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_id", schema2.arrow)


def test_update_success(schema_a, schema_b):
    schema = Schema.of({"key": schema_a.arrow})
    new_schema = schema.replace_subschema("key", schema_b.arrow)
    assert schema.subschemas["key"].equivalent_to(schema_a)
    assert new_schema.subschemas["key"].equivalent_to(schema_b)


def test_update_not_exist(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.replace_subschema("nonexistent", schema_a.arrow)


def test_delete_schema_success(schema_a, schema_b):
    schema = Schema.of({"key1": schema_a.arrow, "key2": schema_b.arrow})
    new_schema = schema.delete_subschema("key1")
    assert "key1" in schema.subschemas
    assert "key2" in schema.subschemas
    assert "key1" not in new_schema.subschemas
    assert "key2" in new_schema.subschemas
    assert "key1" in schema.subschemas_to_field_ids
    assert "key2" in schema.subschemas_to_field_ids
    assert "key1" not in new_schema.subschemas_to_field_ids
    assert "key2" in new_schema.subschemas_to_field_ids


def test_delete_only_schema_fails(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.delete_subschema("key")


def test_delete_schema_not_exist(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.delete_subschema("nonexistent")


def test_get_schemas_order(schema_a, schema_b, schema_c):
    schema = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow, "c": schema_c.arrow})
    assert list(schema.subschemas.keys()) == ["a", "b", "c"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b, schema_c]
    schema = Schema.of({"a": schema_a.arrow})
    schema = schema.add_subschema("b", schema_b.arrow)
    assert list(schema.subschemas.keys()) == ["a", "b"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b]
    schema = schema.add_subschema("c", schema_c.arrow)
    assert list(schema.subschemas.keys()) == ["a", "b", "c"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b, schema_c]


def test_equivalent_to_same(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    assert schema1.equivalent_to(schema2)
    assert schema2.equivalent_to(schema1)


def test_equivalent_to_different_subschema_names(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"x": schema_a.arrow, "b": schema_b.arrow})
    assert schema1.equivalent_to(schema2)
    assert not schema1.equivalent_to(schema2, True)


def test_not_equivalent_to_different_subschema_order(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"b": schema_b.arrow, "a": schema_a.arrow})
    assert not schema1.equivalent_to(schema2)


def test_equivalent_to_non_schema(schema_a):
    schema = Schema.of({"a": schema_a.arrow})
    assert not schema.equivalent_to("not a schema")


def test_equivalent_schemas_different_instances():
    """
    Edge case: ensure equivalent schemas with different instances
    are considered equivalent.
    """
    schema1 = Schema.of(
        [
            Field.of(
                pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema1 = Schema.of({"key": schema1.arrow})
    schema2 = Schema.of({"key": schema2.arrow})
    assert schema1.equivalent_to(schema2)


def test_empty_schema_fails():
    with pytest.raises(ValueError):
        Schema.of({})
    with pytest.raises(ValueError):
        Schema.of([])


def test_schema_type_promotion_edge_cases():
    """Test edge cases for type promotion with SchemaConsistencyType.NONE."""
    # Test 1: Same type - no promotion
    field_int32 = Field.of(
        pa.field("test", pa.int32()), consistency_type=SchemaConsistencyType.NONE
    )
    data_int32 = pa.array([1, 2, 3], type=pa.int32())
    promoted_data, was_promoted = field_int32.promote_type_if_needed(data_int32)
    assert not was_promoted, "Same type should not trigger promotion"
    assert promoted_data.type == pa.int32(), "Data type should remain int32"

    # Test 2: int32 to int64 promotion
    data_int64 = pa.array([2147483648], type=pa.int64())  # Value requiring int64
    promoted_data, was_promoted = field_int32.promote_type_if_needed(data_int64)
    assert was_promoted, "int32 field should promote to int64"
    assert promoted_data.type == pa.int64(), "Promoted data should be int64"

    # Test 3: Nullability preservation
    field_nullable = Field.of(
        pa.field("test", pa.int32(), nullable=True),
        consistency_type=SchemaConsistencyType.NONE,
    )
    data_with_null = pa.array([1, None, 3], type=pa.int32())
    promoted_data, was_promoted = field_nullable.promote_type_if_needed(data_with_null)
    assert not was_promoted, "Same nullable type should not promote"

    # Test 4: Cross-type promotion (int to float)
    field_int = Field.of(
        pa.field("test", pa.int32()), consistency_type=SchemaConsistencyType.NONE
    )
    data_float = pa.array([1.5, 2.7], type=pa.float64())
    promoted_data, was_promoted = field_int.promote_type_if_needed(data_float)
    assert was_promoted, "int32 should promote to accommodate float64"
    assert pa.types.is_floating(
        promoted_data.type
    ), f"Should promote to float type, got {promoted_data.type}"


def test_schema_update_method(schema_a):
    """Test the Schema.update() convenience method."""
    # Test basic usage
    update = schema_a.update()
    assert isinstance(update, SchemaUpdate)
    assert update.base_schema == schema_a
    assert not update.allow_incompatible_changes

    # Test with allow_incompatible_changes=True
    update_permissive = schema_a.update(allow_incompatible_changes=True)
    assert isinstance(update_permissive, SchemaUpdate)
    assert update_permissive.base_schema == schema_a
    assert update_permissive.allow_incompatible_changes

    # Test method chaining with field addition
    new_field = Field.of(pa.field("name", pa.string(), nullable=True), field_id=4)
    updated_schema = schema_a.update().add_field(new_field).apply()

    assert len(updated_schema.fields) == 2
    assert updated_schema.field("col1") == schema_a.field(
        "col1"
    )  # Original field preserved
    added_field = updated_schema.field("name")
    assert added_field.arrow.name == "name"
    assert added_field.arrow.type == pa.string()
    assert added_field.id == 2  # requested field_id of 4 is ignored and auto-assigned


def test_default_value_type_promotion():
    """Test that default values are correctly cast when field types are promoted."""

    # Test 1: Unit-level default value casting
    # Create a field with int32 type and default values
    original_field = Field.of(
        pa.field("test_field", pa.int32()),
        past_default=42,
        future_default=100,
        consistency_type=SchemaConsistencyType.NONE,
    )

    # Test casting to int64
    promoted_past = original_field._cast_default_to_promoted_type(42, pa.int64())
    promoted_future = original_field._cast_default_to_promoted_type(100, pa.int64())
    assert promoted_past == 42
    assert promoted_future == 100

    # Test casting to float64
    promoted_past_float = original_field._cast_default_to_promoted_type(
        42, pa.float64()
    )
    promoted_future_float = original_field._cast_default_to_promoted_type(
        100, pa.float64()
    )
    assert promoted_past_float == 42.0
    assert promoted_future_float == 100.0

    # Test casting to string
    promoted_past_str = original_field._cast_default_to_promoted_type(42, pa.string())
    promoted_future_str = original_field._cast_default_to_promoted_type(
        100, pa.string()
    )
    assert promoted_past_str == "42"
    assert promoted_future_str == "100"

    # Test 2: Test that the default casting logic works correctly
    # Test with None values (should return None)
    none_result = original_field._cast_default_to_promoted_type(None, pa.string())
    assert none_result is None, "None default should remain None"

    # Test error handling - incompatible cast should raise SchemaValidationError
    with pytest.raises(SchemaValidationError):
        original_field._cast_default_to_promoted_type("not_a_number", pa.int64())

    # Test with a complex type
    complex_field = Field.of(
        pa.field("complex", pa.list_(pa.int32())),
        consistency_type=SchemaConsistencyType.NONE,
    )
    with pytest.raises(SchemaValidationError):
        complex_field._cast_default_to_promoted_type(42, pa.list_(pa.string()))


def test_default_value_backfill_with_promotion():
    """Test that default values are correctly backfilled when types are promoted."""

    # Test the interaction between default value casting and binary promotion
    # This represents a common scenario where defaults need to be promoted to binary
    field_with_defaults = Field.of(
        pa.field("test_field", pa.int32()),
        past_default=42,
        future_default=100,
        consistency_type=SchemaConsistencyType.NONE,
    )

    # Test promotion to string (a common "catch-all" type in type promotion)
    string_past = field_with_defaults._cast_default_to_promoted_type(42, pa.string())
    string_future = field_with_defaults._cast_default_to_promoted_type(100, pa.string())

    assert string_past == "42", f"Expected '42', got {string_past}"
    assert string_future == "100", f"Expected '100', got {string_future}"

    # Also test floats to string
    float_field = Field.of(
        pa.field("float_field", pa.float32()),
        past_default=3.14159,
        future_default=2.71828,
        consistency_type=SchemaConsistencyType.NONE,
    )

    string_past = float_field._cast_default_to_promoted_type(3.14159, pa.string())
    string_future = float_field._cast_default_to_promoted_type(2.71828, pa.string())

    assert string_past == "3.14159", f"Expected '3.14159', got {string_past}"
    assert string_future == "2.71828", f"Expected '2.71828', got {string_future}"

    # Test that None defaults are handled correctly
    none_field = Field.of(
        pa.field("none_field", pa.int32()),
        past_default=None,
        future_default=42,
        consistency_type=SchemaConsistencyType.NONE,
    )

    none_past = none_field._cast_default_to_promoted_type(None, pa.string())
    valid_future = none_field._cast_default_to_promoted_type(42, pa.string())

    assert none_past is None, f"None should remain None, got {none_past}"
    assert valid_future == "42", f"Expected '42', got {valid_future}"



from __future__ import annotations

from typing import Optional

from pyarrow import RecordBatch

from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider
from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet.reader.data_reader import (
    RowAndKey,
    FileReader,
    FILE_FORMAT,
)
from deltacat.experimental.storage.rivulet.reader.pyarrow_data_reader import (
    RecordBatchRowIndex,
)
from deltacat.experimental.storage.rivulet.schema.schema import Schema
import pyarrow.parquet as pq
import pyarrow as pa


class ParquetFileReader(FileReader[RecordBatchRowIndex]):
    """
    Parquet file reader

    This class is not thread safe
    """

    def __init__(
        self,
        sst_row: SSTableRow,
        file_provider: FileProvider,
        key: str,
        schema: Schema,
        iter_batch_size=1000,
    ):
        self.sst_row = sst_row
        self.input = file_provider.provide_input_file(sst_row.uri)

        self.key = key
        self.parquet_file: pa.parquet.ParquetFile | None = None
        self.iter_batch_size = iter_batch_size

        # Iterator from pyarrow iter_batches API call. Pyarrow manages state of traversal within parquet row groups
        self._record_batch_iter = None

        self.schema = schema

        """
        These variables keep state about where the iterator is current at. They are initialized in __enter__()
        """
        self._curr_batch: RecordBatch | None = None
        self._curr_row_offset = 0
        self._pk_col = None

    def peek(self) -> Optional[RowAndKey[FILE_FORMAT]]:
        """
        Peek next record

        Note that there is an edge case where peek() is called on the bounary between record batches
        This only happens curr_row_offset == curr_batch.num_rows, meaning next() or peek() would need to advance
        to the next record batch. When this happens, peek() increments _curr_batch and sets _curr_row_offset to 0

        :return: Optional of RowAndKey
        """
        if not self.__is_initialized():
            raise RuntimeError(
                "ParquetFileReader must be initialized with __enter__ before reading"
            )

        if self.__need_to_advance_record_batch():
            try:
                self.__advance_record_batch()
            except StopIteration:
                return None

        pk = self._pk_col[self._curr_row_offset].as_py()
        return RowAndKey(
            RecordBatchRowIndex(self._curr_batch, self._curr_row_offset), pk
        )

    def __next__(self) -> RowAndKey[FILE_FORMAT]:
        if not self.__is_initialized():
            raise RuntimeError(
                "ParquetFileReader must be initialized with __enter__ before reading"
            )

        if self.__need_to_advance_record_batch():
            self.__advance_record_batch()
            pk = self._pk_col[0].as_py()
            return RowAndKey(RecordBatchRowIndex(self._curr_batch, 0), pk)
        else:
            pk = self._pk_col[self._curr_row_offset].as_py()
            offset = self._curr_row_offset
            self._curr_row_offset += 1
            return RowAndKey(RecordBatchRowIndex(self._curr_batch, offset), pk)

    def __enter__(self):
        with self.input.open() as f:
            self.parquet_file = pq.ParquetFile(f)
            # Initialize _curr_batch
            row_groups = list(range(self.sst_row.offset_start, self.sst_row.offset_end))
            self._record_batch_iter = self.parquet_file.iter_batches(
                self.iter_batch_size, row_groups, columns=self.schema.keys()
            )
            self.__advance_record_batch()

    def __exit__(self, __exc_type, __exc_value, __traceback):
        self.close()
        # return False to propagate up error messages
        return False

    def close(self):
        self.parquet_file.close()

    def __is_initialized(self):
        return self.parquet_file and self._curr_batch and self._pk_col

    def __need_to_advance_record_batch(self):
        return not self._curr_row_offset < self._curr_batch.num_rows

    def __advance_record_batch(self):
        """
        Advance to next record batch
        :raise StopIteration: If there are no more record batches
        """
        self._curr_batch = next(self._record_batch_iter)
        self._curr_row_offset = 0
        self._pk_col = self._curr_batch[self.key]



from tenacity import (
    Retrying,
    retry_if_exception_type,
    stop_after_delay,
    wait_random_exponential,
)
from typing import Union, Optional, Dict, Any, List, Callable
from deltacat.types.tables import (
    CapturedBlockWritePaths,
    UuidBlockWritePathProvider,
)
from deltacat.types.tables import (
    get_table_writer,
    get_table_length,
    TABLE_CLASS_TO_SLICER_FUNC,
)
from deltacat.exceptions import RetryableError
from deltacat.storage import (
    DistributedDataset,
    LocalTable,
)
from deltacat.types.media import (
    ContentEncoding,
    ContentType,
)
from deltacat.constants import UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY
import s3fs
import boto3
from boto3.session import Session
from botocore.credentials import Credentials


def get_credential() -> Credentials:
    boto3_session: Session = boto3.Session()
    credentials: Credentials = boto3_session.get_credentials()
    return credentials


def get_s3_file_system(content_type):
    import s3fs  # noqa: F401

    token_holder = get_credential()
    content_encoding = ContentEncoding.IDENTITY

    s3_file_system = s3fs.S3FileSystem(
        key=token_holder.access_key,
        secret=token_holder.secret_key,
        token=token_holder.token,
        s3_additional_kwargs={
            "ServerSideEncryption": "aws:kms",
            # TODO: Get tagging from table properties
            "ContentType": content_type.value,
            "ContentEncoding": content_encoding.value,
        },
    )
    return s3_file_system


def upload_table_with_retry(
    table: Union[LocalTable, DistributedDataset],
    s3_url_prefix: str,
    s3_table_writer_kwargs: Optional[Dict[str, Any]],
    content_type: ContentType = ContentType.PARQUET,
    max_records_per_file: Optional[int] = 4000000,
    filesystem: Optional[s3fs.S3FileSystem] = None,
    **s3_client_kwargs: Any,
) -> List[str]:
    """
    Writes the given table to 1 or more S3 files and return the paths
    of the S3 files written.
    """
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )

    if s3_table_writer_kwargs is None:
        s3_table_writer_kwargs = {}

    if not filesystem:
        filesystem = get_s3_file_system(content_type=content_type)
    capture_object = CapturedBlockWritePaths()
    block_write_path_provider = UuidBlockWritePathProvider(
        capture_object=capture_object, base_path=s3_url_prefix
    )
    s3_table_writer_func = get_table_writer(table)
    table_record_count = get_table_length(table)
    if max_records_per_file is None or not table_record_count:
        retrying(
            fn=upload_table,
            table_slices=table,
            s3_base_url=f"{s3_url_prefix}",
            s3_file_system=filesystem,
            s3_table_writer_func=s3_table_writer_func,
            s3_table_writer_kwargs=s3_table_writer_kwargs,
            block_write_path_provider=block_write_path_provider,
            content_type=content_type,
            **s3_client_kwargs,
        )
    else:
        table_slicer_func = TABLE_CLASS_TO_SLICER_FUNC.get(type(table))
        table_slices = table_slicer_func(table, max_records_per_file)
        for table_slice in table_slices:
            retrying(
                fn=upload_table,
                table_slices=table_slice,
                s3_base_url=f"{s3_url_prefix}",
                s3_file_system=filesystem,
                s3_table_writer_func=s3_table_writer_func,
                s3_table_writer_kwargs=s3_table_writer_kwargs,
                block_write_path_provider=block_write_path_provider,
                content_type=content_type,
                **s3_client_kwargs,
            )
    del block_write_path_provider
    write_paths = capture_object.write_paths()
    s3_write_paths = []
    for path in write_paths:
        s3_write_path = construct_s3_url(path)
        s3_write_paths.append(s3_write_path)
    return s3_write_paths


def construct_s3_url(path: Optional[str]) -> Optional[str]:
    if path:
        return f"s3://{path}"
    return None


def upload_table(
    table_slices: Union[LocalTable, DistributedDataset],
    s3_base_url: str,
    s3_file_system: s3fs.S3FileSystem,
    s3_table_writer_func: Callable,
    block_write_path_provider: UuidBlockWritePathProvider,
    content_type: ContentType,
    s3_table_writer_kwargs: Dict[str, Any],
) -> None:
    s3_table_writer_func(
        table_slices,
        s3_base_url,
        s3_file_system,
        block_write_path_provider,
        content_type.value,
        **s3_table_writer_kwargs,
    )
    # TODO: Add a proper fix for block_refs and write_paths not persisting in Ray actors



from dataclasses import dataclass, fields

from deltacat.storage import (
    Partition,
)
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from typing import Optional


@dataclass(frozen=True)
class ExecutionCompactionResult:
    new_compacted_partition: Optional[Partition]
    new_round_completion_info: Optional[RoundCompletionInfo]
    is_inplace_compacted: bool

    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))



import pytest
import pyarrow as pa
from daft import DataType, TimeUnit
from daft.logical.schema import Field as DaftField

from deltacat.storage.model.transform import IdentityTransform
from deltacat.storage.model.partition import PartitionKey
from deltacat.utils.daft import DaftFieldMapper, DaftPartitionKeyMapper

from deltacat.storage.model.schema import Field, Schema


class TestDaftFieldMapper:
    def test_field_mapper_basic_types(self):
        """Test mapping basic data types between Daft and PyArrow fields"""
        test_cases = [
            (DataType.int32(), pa.int32()),
            (DataType.int64(), pa.int64()),
            (DataType.float32(), pa.float32()),
            (DataType.float64(), pa.float64()),
            (DataType.string(), pa.large_string()),
            (DataType.bool(), pa.bool_()),
            (DataType.binary(), pa.large_binary()),
            (DataType.date(), pa.date32()),
            (DataType.timestamp(TimeUnit.ns()), pa.timestamp("ns")),
        ]

        for daft_type, pa_type in test_cases:
            # Create test fields
            daft_field = DaftField.create(
                name="test_field",
                dtype=daft_type,
            )

            # Daft to PyArrow
            pa_field = DaftFieldMapper.map(daft_field)
            assert pa_field is not None
            assert pa_field.name == "test_field"
            assert pa_field.type == pa_type  # type: ignore
            assert pa_field.nullable is True

            # PyArrow to Daft
            daft_field_back = DaftFieldMapper.unmap(pa_field)
            assert daft_field_back is not None
            assert daft_field_back.name == daft_field.name
            assert daft_field_back.dtype == daft_field.dtype


class TestDaftPartitionKeyMapper:
    def test_unmap(self):
        """
        Test unmap method of DaftPartitionKeyMapper when obj is not None, schema is provided,
        len(obj.key) is 1, and dc_field is found in the schema.

        This test verifies that the method correctly converts a PartitionKey to a DaftPartitionField
        when all conditions are met and the field exists in the schema.
        """
        # Create a mock schema
        schema = Schema.of(schema=[Field.of(pa.field("test_field", pa.int32()))])
        # Create a PartitionKey object
        partition_key = PartitionKey(
            key=["test_field"], transform=IdentityTransform(), name="partition_field"
        )

        result = DaftPartitionKeyMapper.unmap(obj=partition_key, schema=schema)
        assert result is not None
        assert result.field.name() == "partition_field"
        assert DataType._from_pydatatype(result.field.dtype()) == DataType.int32()

    def test_unmap_no_field_locator(self):
        schema = Schema.of(schema=[Field.of(pa.field("test_field", pa.int32()))])
        partition_key = PartitionKey(key=[], name="partition_field")

        with pytest.raises(ValueError) as excinfo:
            DaftPartitionKeyMapper.unmap(partition_key, schema)

        assert "At least 1 PartitionKey FieldLocator is expected" in str(excinfo.value)

    def test_unmap_partition_key_not_found(self):
        schema = Schema.of(schema=[Field.of(pa.field("test_field", pa.int32()))])
        partition_key = PartitionKey(
            key=["test_field_2"], transform=IdentityTransform(), name="partition_field"
        )

        with pytest.raises(KeyError) as excinfo:
            DaftPartitionKeyMapper.unmap(partition_key, schema)

        assert "Column test_field_2 does not exist in schema" in str(excinfo.value)

    def test_unmap_partition_name_not_defined(self):
        schema = Schema.of(schema=[Field.of(pa.field("test_field", pa.int32()))])
        partition_key = PartitionKey(key=[])

        with pytest.raises(ValueError) as excinfo:
            DaftPartitionKeyMapper.unmap(partition_key, schema)

        assert "Name is required for PartitionKey conversion" in str(excinfo.value)



from __future__ import annotations

from enum import Enum
from typing import List, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import polars as pl
from ray.data.dataset import Dataset as RayDataset
from daft import DataFrame as DaftDataFrame

from deltacat.constants import (
    RUNNING_TXN_DIR_NAME,
    PAUSED_TXN_DIR_NAME,
    FAILED_TXN_DIR_NAME,
    SUCCESS_TXN_DIR_NAME,
)

LocalTable = Union[
    pa.Table,
    pd.DataFrame,
    pl.DataFrame,
    np.ndarray,
    pa.parquet.ParquetFile,
]
LocalDataset = Union[LocalTable, List[LocalTable]]
DistributedDataset = Union[RayDataset, DaftDataFrame]
Dataset = Union[LocalDataset, DistributedDataset]


class StreamFormat(str, Enum):
    DELTACAT = "deltacat"
    ICEBERG = "iceberg"
    HIVE = "hive"
    HUDI = "hudi"
    DELTA_LAKE = "delta_lake"
    SQLITE3 = "SQLITE3"  # used by tests


class DeltaType(str, Enum):
    APPEND = "append"
    UPSERT = "upsert"
    DELETE = "delete"


class TransactionOperationType(str, Enum):
    CREATE = "create"
    UPDATE = "update"
    REPLACE = "replace"
    DELETE = "delete"

    READ_SIBLINGS = "read_siblings"
    READ_CHILDREN = "read_children"
    READ_LATEST = "read_latest"
    READ_EXISTS = "read_exists"

    @staticmethod
    def write_operations():
        return {
            TransactionOperationType.CREATE,
            TransactionOperationType.UPDATE,
            TransactionOperationType.REPLACE,
            TransactionOperationType.DELETE,
        }

    @staticmethod
    def read_operations():
        return {
            TransactionOperationType.READ_SIBLINGS,
            TransactionOperationType.READ_CHILDREN,
            TransactionOperationType.READ_LATEST,
            TransactionOperationType.READ_EXISTS,
        }

    def is_write_operation(self) -> bool:
        return self in TransactionOperationType.write_operations()

    def is_read_operation(self) -> bool:
        return self in TransactionOperationType.read_operations()


class TransactionStatus(str, Enum):
    """
    Transaction user status types. Every transaction status maps to a distinct
    transaction log directory.
    """

    SUCCESS = "SUCCESS"
    RUNNING = "RUNNING"
    PAUSED = "PAUSED"
    FAILED = "FAILED"

    def dir_name(self) -> str:
        if self == TransactionStatus.RUNNING:
            return RUNNING_TXN_DIR_NAME
        elif self == TransactionStatus.PAUSED:
            return PAUSED_TXN_DIR_NAME
        elif self == TransactionStatus.FAILED:
            return FAILED_TXN_DIR_NAME
        elif self == TransactionStatus.SUCCESS:
            return SUCCESS_TXN_DIR_NAME


class TransactionState(str, Enum):
    """
    Transaction system state types. Transaction states do not map to distinct transaction log directories,
    but can be inferred by its presence in one or more directories. These states are used to infer whether
    to run system activities like transaction cleanup jobs.
    """

    FAILED = "FAILED"
    PURGED = "PURGED"
    TIMEOUT = "TIMEOUT"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    PAUSED = "PAUSED"


class LifecycleState(str, Enum):
    CREATED = "created"
    UNRELEASED = "unreleased"
    ACTIVE = "active"
    DEPRECATED = "deprecated"
    BETA = "beta"
    DELETED = "deleted"


class CommitState(str, Enum):
    STAGED = "staged"
    COMMITTED = "committed"
    DEPRECATED = "deprecated"


class SchemaConsistencyType(str, Enum):
    """
    DeltaCAT table schemas can be used to inform the data consistency checks
    run for each field. When present, the schema can be used to enforce the
    following field-level data consistency policies at table load time:

    NONE: No consistency checks are run.

    COERCE: Coerce fields to fit the schema whenever possible.

    VALIDATE: Raise an error for any fields that don't fit the schema.
    """

    NONE = "none"
    COERCE = "coerce"
    VALIDATE = "validate"


class SortOrder(str, Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"

    @classmethod
    def _missing_(cls, value: str):
        # pyiceberg.table.sorting.SortDirection mappings
        if value.lower() == "asc":
            return SortOrder.ASCENDING
        elif value.lower() == "desc":
            return SortOrder.DESCENDING
        return None


class NullOrder(str, Enum):
    AT_START = "at_start"
    AT_END = "at_end"

    @classmethod
    def _missing_(cls, value: str):
        # pyiceberg.table.sorting.NullOrder mappings
        if value.lower() == "nulls-first":
            return NullOrder.AT_START
        elif value.lower() == "nulls-last":
            return NullOrder.AT_END
        return None



import logging
from typing import Optional, Dict, Any, List
from deltacat import logs
from deltacat.storage import (
    Delta,
    interface as unimplemented_deltacat_storage,
)
from deltacat.compute.compactor_v2.utils.content_type_params import (
    append_content_type_params,
)
from deltacat.compute.resource_estimation.model import (
    OperationType,
    EstimateResourcesParams,
    ResourceEstimationMethod,
    EstimatedResources,
    Statistics,
)
from deltacat.compute.resource_estimation.manifest import (
    estimate_manifest_entry_size_bytes,
    estimate_manifest_entry_num_rows,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _estimate_resources_required_to_process_delta_using_previous_inflation(
    delta: Delta,
    operation_type: OperationType,
    estimate_resources_params: EstimateResourcesParams,
    deltacat_storage: unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Dict[str, Any],
    **kwargs,
) -> Optional[EstimatedResources]:

    assert (
        operation_type == OperationType.PYARROW_DOWNLOAD
    ), "Number of rows can only be estimated for PYARROW_DOWNLOAD operation"
    assert (
        estimate_resources_params.previous_inflation
    ), "Previous inflation must be provided to estimate delta size"

    in_memory_size = (
        delta.meta.content_length * estimate_resources_params.previous_inflation
    )
    num_rows = 0
    if estimate_resources_params.average_record_size_bytes is not None:
        num_rows = int(
            in_memory_size / estimate_resources_params.average_record_size_bytes
        )

    return EstimatedResources.of(
        memory_bytes=in_memory_size,
        statistics=Statistics.of(
            in_memory_size_bytes=in_memory_size,
            record_count=num_rows,
            on_disk_size_bytes=delta.meta.content_length,
        ),
    )


def _estimate_resources_required_to_process_delta_using_type_params(
    delta: Delta,
    operation_type: OperationType,
    all_column_names: List[str],
    estimate_resources_params: EstimateResourcesParams,
    deltacat_storage: unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Dict[str, Any],
    **kwargs,
) -> Optional[EstimatedResources]:

    assert (
        operation_type == OperationType.PYARROW_DOWNLOAD
    ), "Number of rows can only be estimated for PYARROW_DOWNLOAD operation"

    if estimate_resources_params.parquet_to_pyarrow_inflation is None:
        logger.debug(
            "Could not estimate using type params as "
            f"parquet_to_pyarrow_inflation is None for {delta.locator}"
        )
        return None

    if not delta.manifest:
        delta.manifest = deltacat_storage.get_delta_manifest(
            delta.locator,
            **deltacat_storage_kwargs,
        )

    if not delta.manifest or not delta.manifest.entries:
        return EstimatedResources.of(
            memory_bytes=0,
            statistics=Statistics.of(
                in_memory_size_bytes=0,
                record_count=0,
                on_disk_size_bytes=delta.meta.content_length,
            ),
        )
    file_reader_kwargs_provider = kwargs.get(
        "file_reader_kwargs_provider"
    ) or deltacat_storage_kwargs.get("file_reader_kwargs_provider")

    """
    NOTE: The file_reader_kwargs_provider parameter can be passed in two ways:
    1. Nested within deltacat_storage_kwargs during resource estimation
    2. As a top-level attribute of CompactPartitionsParams during compaction

    This creates an inconsistent parameter path between resource estimation and compaction flows.
    As a long-term solution, this should be unified to use a single consistent path (either always
    nested in deltacat_storage_kwargs or always as a top-level parameter).

    For now, this implementation handles the resource estimation case by:
    1. First checking for file_reader_kwargs_provider as a direct kwarg
    2. Falling back to deltacat_storage_kwargs if not found
    This approach maintains backward compatibility by not modifying the DELTA_RESOURCE_ESTIMATION_FUNCTIONS signatures.
    """
    appended = append_content_type_params(
        delta=delta,
        all_column_names=all_column_names,
        deltacat_storage=deltacat_storage,
        deltacat_storage_kwargs=deltacat_storage_kwargs,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
    )

    if not appended:
        logger.debug(
            f"Could not append content type params for {delta.locator}, returning None"
        )
        return None

    in_memory_size = 0.0
    num_rows = 0

    for entry in delta.manifest.entries:
        cur_memory = estimate_manifest_entry_size_bytes(
            entry=entry,
            operation_type=operation_type,
            estimate_resources_params=estimate_resources_params,
            **kwargs,
        )
        cur_num_rows = estimate_manifest_entry_num_rows(
            entry=entry,
            operation_type=operation_type,
            estimate_resources_params=estimate_resources_params,
            **kwargs,
        )

        if cur_memory is None or cur_num_rows is None:
            return None

        in_memory_size += cur_memory
        num_rows += cur_num_rows

    return EstimatedResources.of(
        memory_bytes=in_memory_size,
        statistics=Statistics.of(
            in_memory_size_bytes=in_memory_size,
            record_count=num_rows,
            on_disk_size_bytes=delta.meta.content_length,
        ),
    )


def _estimate_resources_required_to_process_delta_using_file_sampling(
    delta: Delta,
    operation_type: OperationType,
    estimate_resources_params: EstimateResourcesParams,
    deltacat_storage: unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Dict[str, Any],
    **kwargs,
) -> Optional[EstimatedResources]:

    assert (
        operation_type == OperationType.PYARROW_DOWNLOAD
    ), "Number of rows can only be estimated for PYARROW_DOWNLOAD operation"

    if not estimate_resources_params.max_files_to_sample:
        # we cannot calculate if we cannot sample
        return None

    if not delta.manifest:
        delta.manifest = deltacat_storage.get_delta_manifest(
            delta.locator,
            **deltacat_storage_kwargs,
        )

    if not delta.manifest or not delta.manifest.entries:
        return EstimatedResources.of(
            memory_bytes=0,
            statistics=Statistics.of(
                in_memory_size_bytes=0,
                record_count=0,
                on_disk_size_bytes=delta.meta.content_length,
            ),
        )

    sampled_in_memory_size = 0.0
    sampled_on_disk_size = 0.0
    sampled_num_rows = 0

    for entry_index in range(
        min(estimate_resources_params.max_files_to_sample, len(delta.manifest.entries))
    ):
        tbl = deltacat_storage.download_delta_manifest_entry(
            delta,
            entry_index,
            **deltacat_storage_kwargs,
        )
        sampled_in_memory_size += tbl.nbytes
        sampled_on_disk_size += delta.manifest.entries[entry_index].meta.content_length
        sampled_num_rows += len(tbl)

    if not sampled_on_disk_size or not sampled_in_memory_size:
        return EstimatedResources.of(
            memory_bytes=0,
            statistics=Statistics.of(
                in_memory_size_bytes=0,
                record_count=0,
                on_disk_size_bytes=delta.meta.content_length,
            ),
        )

    sampled_inflation = sampled_in_memory_size / sampled_on_disk_size

    in_memory_size = sampled_inflation * delta.meta.content_length
    num_rows = int(in_memory_size / sampled_in_memory_size * sampled_num_rows)

    return EstimatedResources.of(
        memory_bytes=in_memory_size,
        statistics=Statistics.of(
            in_memory_size_bytes=in_memory_size,
            record_count=num_rows,
            on_disk_size_bytes=delta.meta.content_length,
        ),
    )


RESOURCE_ESTIMATION_METHOD_TO_DELTA_RESOURCE_ESTIMATION_FUNCTIONS = {
    ResourceEstimationMethod.PREVIOUS_INFLATION: [
        _estimate_resources_required_to_process_delta_using_previous_inflation
    ],
    ResourceEstimationMethod.CONTENT_TYPE_META: [
        _estimate_resources_required_to_process_delta_using_type_params
    ],
    ResourceEstimationMethod.INTELLIGENT_ESTIMATION: [
        _estimate_resources_required_to_process_delta_using_type_params,
    ],
    ResourceEstimationMethod.FILE_SAMPLING: [
        _estimate_resources_required_to_process_delta_using_file_sampling
    ],
    ResourceEstimationMethod.DEFAULT: [
        _estimate_resources_required_to_process_delta_using_previous_inflation,
    ],
    ResourceEstimationMethod.DEFAULT_V2: [
        _estimate_resources_required_to_process_delta_using_type_params,
        _estimate_resources_required_to_process_delta_using_file_sampling,
        _estimate_resources_required_to_process_delta_using_previous_inflation,
    ],
    ResourceEstimationMethod.FILE_SAMPLING_WITH_PREVIOUS_INFLATION: [
        _estimate_resources_required_to_process_delta_using_file_sampling,
        _estimate_resources_required_to_process_delta_using_previous_inflation,
    ],
}


def estimate_resources_required_to_process_delta(
    delta: Delta,
    operation_type: OperationType,
    estimate_resources_params: EstimateResourcesParams = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Dict[str, Any] = {},
    **kwargs,
) -> Optional[EstimatedResources]:
    assert (
        operation_type == OperationType.PYARROW_DOWNLOAD
    ), "Number of rows can only be estimated for PYARROW_DOWNLOAD operation"

    if delta.meta.record_count and delta.meta.source_content_length:
        # No need to estimate
        return EstimatedResources.of(
            memory_bytes=delta.meta.source_content_length,
            statistics=Statistics.of(
                in_memory_size_bytes=delta.meta.source_content_length,
                record_count=delta.meta.record_count,
                on_disk_size_bytes=delta.meta.content_length,
            ),
        )

    if estimate_resources_params is None:
        estimate_resources_params = EstimateResourcesParams.of()

    functions = RESOURCE_ESTIMATION_METHOD_TO_DELTA_RESOURCE_ESTIMATION_FUNCTIONS.get(
        estimate_resources_params.resource_estimation_method
    )

    all_column_names = deltacat_storage.get_table_version_column_names(
        delta.locator.namespace,
        delta.locator.table_name,
        delta.locator.table_version,
        **deltacat_storage_kwargs,
    )

    for func in functions:
        resources = func(
            delta=delta,
            operation_type=operation_type,
            all_column_names=all_column_names,
            estimate_resources_params=estimate_resources_params,
            deltacat_storage=deltacat_storage,
            deltacat_storage_kwargs=deltacat_storage_kwargs,
            **kwargs,
        )
        if resources is not None:
            logger.debug(
                f"Estimated resources for delta={delta.locator} is {resources} using {func}"
            )
            return resources

    return None



import sys
import time
from contextlib import contextmanager
from typing import Generator, Tuple

from deltacat.benchmarking.benchmark_report import BenchmarkMetric, BenchmarkStep
from deltacat.experimental.storage.rivulet.dataset import Dataset
from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)


@contextmanager
def timed_step(description: str) -> Generator[BenchmarkStep, None, None]:
    """Convenience for computing elapsed time of a block of code as a metric.

    :param description: description of the step
    :return: a benchmark operation populated with the elapsed time
    """
    metric = BenchmarkStep(description)
    start_time = time.time()
    yield metric
    end_time = time.time()
    metric.add(BenchmarkMetric("elapsed_time", 1000 * (end_time - start_time), "ms"))


class BenchmarkEngine:
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def load_and_commit(
        self, schema_name, generator, count
    ) -> Tuple[str, BenchmarkStep]:
        """Load count number of rows from the generator and commit.

        :param generator: row generator
        :param count: the number of rows to load into the dataset
        :return: tuple of the manifest URI and a operation measurement
        """
        desc = f"load {count} from {generator}"
        writer = self.dataset.writer(schema_name)
        with timed_step(desc) as step:
            rows = [generator.generate() for _ in range(count)]
            writer.write(rows)
            result = writer.flush()
        step.add(BenchmarkMetric("loaded", count))
        return result, step

    def scan(self) -> Tuple[set[any], BenchmarkStep]:
        """
        Scans the rows in dataset and prints some basic statistics about the manifest

        :return: Tuple[set[any], BenchmarkStep] - a tuple containing a set of merge keys and a benchmark step with metrics
        """
        keys = set()
        object_count = 0
        size_b = 0
        # Note that we expect single col merge keys so we can return key set
        # this will fail with validation error if dataset has multiple merge keys
        merge_key_name = self.dataset.schemas["all"].get_merge_key()
        with timed_step("full scan") as step:
            for row in self.dataset.scan(QueryExpression()).to_pydict():
                object_count += 1
                size_b += sum([sys.getsizeof(x) for x in row.values()])
                keys.add(row.get(merge_key_name))
                # TODO replace with the actual metrics we want to measure
        step.add(BenchmarkMetric("rows read", object_count))
        step.add(BenchmarkMetric("size", size_b / (1024 * 1024), "MB"))
        return keys, step

    def run_queries(
        self, description, manifest_uri, queries: list[QueryExpression]
    ) -> BenchmarkStep:
        object_count = 0
        size_b = 0
        with timed_step(description) as step:
            for query in queries:
                for row in self.dataset.scan(query).to_pydict():
                    object_count += 1
                    size_b += sum([sys.getsizeof(x) for x in row.values()])
        # TODO replace with the actual metrics we want to measure
        step.add(BenchmarkMetric("rows read", object_count))
        step.add(BenchmarkMetric("size", size_b / (1024 * 1024), "MB"))
        return step



from __future__ import annotations

import re
from typing import Optional, Tuple, Union, List
from datetime import timedelta
from enum import Enum

import sys
import urllib
import pathlib

import pyarrow as pa
from pyarrow.fs import (
    _resolve_filesystem_and_path,
    FileSelector,
    FileInfo,
    FileType,
    FileSystem,
    FSSpecHandler,
    PyFileSystem,
    GcsFileSystem,
    LocalFileSystem,
    S3FileSystem,
    AzureFileSystem,
    HadoopFileSystem,
)

_LOCAL_SCHEME = "local"


class FilesystemType(str, Enum):
    LOCAL = "local"
    S3 = "s3"
    GCS = "gcs"
    AZURE = "azure"
    HADOOP = "hadoop"
    UNKNOWN = "unknown"

    @classmethod
    def from_filesystem(cls, filesystem: FileSystem) -> FilesystemType:
        if isinstance(filesystem, LocalFileSystem):
            return cls.LOCAL
        elif isinstance(filesystem, S3FileSystem):
            return cls.S3
        elif isinstance(filesystem, GcsFileSystem):
            return cls.GCS
        elif isinstance(filesystem, AzureFileSystem):
            return cls.AZURE
        elif isinstance(filesystem, HadoopFileSystem):
            return cls.HADOOP
        else:
            return cls.UNKNOWN

    @classmethod
    def to_filesystem(cls, filesystem_type: FilesystemType) -> FileSystem:
        if filesystem_type == cls.LOCAL:
            return LocalFileSystem()
        elif filesystem_type == cls.S3:
            return S3FileSystem()
        elif filesystem_type == cls.GCS:
            return GcsFileSystem()
        elif filesystem_type == cls.AZURE:
            return AzureFileSystem()
        elif filesystem_type == cls.HADOOP:
            return HadoopFileSystem()
        else:
            raise ValueError(f"Unsupported filesystem type: {filesystem_type}")


def resolve_paths_and_filesystem(
    paths: Union[str, List[str]],
    filesystem: FileSystem = None,
) -> Tuple[List[str], FileSystem]:
    """
    Resolves and normalizes all provided paths, infers a filesystem from the
    paths or validates the provided filesystem against the paths and ensures
    that all paths use the same filesystem.

    Args:
        paths: A single file/directory path or a list of file/directory paths.
            A list of paths can contain both files and directories.
        filesystem: The filesystem implementation that should be used for
            reading these files. If None, a filesystem will be inferred. If not
            None, the provided filesystem will still be validated against all
            filesystems inferred from the provided paths to ensure
            compatibility.
    """
    if isinstance(paths, str):
        paths = [paths]
    if isinstance(paths, pathlib.Path):
        paths = [str(paths)]
    elif not isinstance(paths, list) or any(not isinstance(p, str) for p in paths):
        raise ValueError(
            "Expected `paths` to be a `str`, `pathlib.Path`, or `list[str]`, but got "
            f"`{paths}`."
        )
    elif len(paths) == 0:
        raise ValueError("Must provide at least one path.")

    need_unwrap_path_protocol = True
    if filesystem and not isinstance(filesystem, FileSystem):
        err_msg = (
            f"The filesystem passed must either conform to "
            f"pyarrow.fs.FileSystem, or "
            f"fsspec.spec.AbstractFileSystem. The provided "
            f"filesystem was: {filesystem}"
        )
        try:
            import fsspec
            from fsspec.implementations.http import HTTPFileSystem
        except ModuleNotFoundError:
            # If filesystem is not a pyarrow filesystem and fsspec isn't
            # installed, then filesystem is neither a pyarrow filesystem nor
            # an fsspec filesystem, so we raise a TypeError.
            raise TypeError(err_msg) from None
        if not isinstance(filesystem, fsspec.spec.AbstractFileSystem):
            raise TypeError(err_msg) from None
        if isinstance(filesystem, HTTPFileSystem):
            # If filesystem is fsspec HTTPFileSystem, the protocol/scheme of paths
            # should not be unwrapped/removed, because HTTPFileSystem expects full file
            # paths including protocol/scheme. This is different behavior compared to
            # file systems implementation in pyarrow.fs.FileSystem.
            need_unwrap_path_protocol = False

        filesystem = PyFileSystem(FSSpecHandler(filesystem))

    resolved_paths = []
    for path in paths:
        path = _resolve_custom_scheme(path)
        try:
            resolved_filesystem, resolved_path = _resolve_filesystem_and_path(
                path, filesystem
            )
        except pa.lib.ArrowInvalid as e:
            if "Cannot parse URI" in str(e):
                resolved_filesystem, resolved_path = _resolve_filesystem_and_path(
                    _encode_url(path), filesystem
                )
                resolved_path = _decode_url(resolved_path)
            elif "Unrecognized filesystem type in URI" in str(e):
                scheme = urllib.parse.urlparse(path, allow_fragments=False).scheme
                if scheme in ["http", "https"]:
                    # If scheme of path is HTTP and filesystem is not resolved,
                    # try to use fsspec HTTPFileSystem. This expects fsspec is
                    # installed.
                    try:
                        from fsspec.implementations.http import HTTPFileSystem
                    except ModuleNotFoundError:
                        raise ImportError(
                            "Please install fsspec to read files from HTTP."
                        ) from None

                    resolved_filesystem = PyFileSystem(FSSpecHandler(HTTPFileSystem()))
                    resolved_path = path
                    need_unwrap_path_protocol = False
                else:
                    raise
            else:
                raise
        if filesystem is None:
            if isinstance(resolved_filesystem, GcsFileSystem):
                # Configure a retry time limit for GcsFileSystem so that it
                # doesn't hang forever trying to get file info (e.g., when
                # trying to get a public file w/o anonymous=True).
                filesystem = GcsFileSystem(
                    retry_time_limit=timedelta(seconds=60),
                )
            else:
                filesystem = resolved_filesystem
        elif need_unwrap_path_protocol:
            resolved_path = _unwrap_protocol(resolved_path)
        resolved_path = filesystem.normalize_path(resolved_path)
        resolved_paths.append(resolved_path)
    return resolved_paths, filesystem


def resolve_path_and_filesystem(
    path: str,
    filesystem: Optional[FileSystem] = None,
) -> Tuple[str, FileSystem]:
    """
    Resolves and normalizes the provided path, infers a filesystem from the
    path or validates the provided filesystem against the path.

    Args:
        path: A single file/directory path.
        filesystem: The filesystem implementation that should be used for
            reading these files. If None, a filesystem will be inferred. If not
            None, the provided filesystem will still be validated against all
            filesystems inferred from the provided paths to ensure
            compatibility.
    """
    paths, filesystem = resolve_paths_and_filesystem(
        paths=path,
        filesystem=filesystem,
    )
    assert len(paths) == 1, len(paths)
    return paths[0], filesystem


def list_directory(
    path: str,
    filesystem: FileSystem,
    exclude_prefixes: Optional[List[str]] = None,
    ignore_missing_path: bool = False,
    recursive: bool = False,
) -> List[Tuple[str, int]]:
    """
    Expand the provided directory path to a list of file paths.

    Args:
        path: The directory path to expand.
        filesystem: The filesystem implementation that should be used for
            reading these files.
        exclude_prefixes: The file relative path prefixes that should be
            excluded from the returned file set. Default excluded prefixes are
            "." and "_".
        recursive: Whether to expand subdirectories or not.

    Returns:
        An iterator of (file_path, file_size) tuples.
    """
    if exclude_prefixes is None:
        exclude_prefixes = [".", "_"]

    selector = FileSelector(
        base_dir=path,
        recursive=recursive,
        allow_not_found=ignore_missing_path,
    )
    try:
        files = filesystem.get_file_info(selector)
    except OSError as e:
        if isinstance(e, FileNotFoundError):
            files = []
        else:
            _handle_read_os_error(e, path)
    base_path = selector.base_dir
    out = []
    for file_ in files:
        file_path = file_.path
        if not file_path.startswith(base_path):
            continue
        relative = file_path[len(base_path) :]
        if any(relative.startswith(prefix) for prefix in exclude_prefixes):
            continue
        out.append((file_path, file_.size))
    # We sort the paths to guarantee a stable order.
    return sorted(out)


def get_file_info(
    path: str,
    filesystem: FileSystem,
    ignore_missing_path: bool = False,
) -> FileInfo:
    """Get the file info for the provided path."""
    try:
        file_info = filesystem.get_file_info(path)
    except OSError as e:
        _handle_read_os_error(e, path)
    if file_info.type == FileType.NotFound and not ignore_missing_path:
        raise FileNotFoundError(path)

    return file_info


def write_file(
    path: str,
    data: Union[str, bytes],
    filesystem: Optional[FileSystem] = None,
) -> None:
    """
    Write data to a file using any filesystem.

    Args:
        path: The file path to write to.
        data: The data to write (string or bytes).
        filesystem: The filesystem implementation to use. If None, will be inferred from the path.
    """
    resolved_path, resolved_filesystem = resolve_path_and_filesystem(
        path=path,
        filesystem=filesystem,
    )

    # Convert string to bytes if necessary
    if isinstance(data, str):
        data = data.encode("utf-8")

    with resolved_filesystem.open_output_stream(resolved_path) as f:
        f.write(data)


def read_file(
    path: str,
    filesystem: Optional[FileSystem] = None,
    fail_if_not_found: bool = True,
) -> Optional[bytes]:
    """
    Read data from a file using any filesystem.

    Args:
        path: The file path to read from.
        filesystem: The filesystem implementation to use. If None, will be inferred from the path.
        fail_if_not_found: Whether to raise an error if the file is not found.

    Returns:
        The file data as bytes, or None if file not found and fail_if_not_found is False.
    """
    try:
        resolved_path, resolved_filesystem = resolve_path_and_filesystem(
            path=path,
            filesystem=filesystem,
        )

        with resolved_filesystem.open_input_stream(resolved_path) as f:
            return f.read()
    except FileNotFoundError:
        if fail_if_not_found:
            raise
        return None


def _handle_read_os_error(
    error: OSError,
    paths: Union[str, List[str]],
) -> str:
    # NOTE: this is not comprehensive yet, and should be extended as more errors arise.
    # NOTE: The latter patterns are raised in Arrow 10+, while the former is raised in
    # Arrow < 10.
    aws_error_pattern = (
        r"^(?:(.*)AWS Error \[code \d+\]: No response body\.(.*))|"
        r"(?:(.*)AWS Error UNKNOWN \(HTTP status 400\) during HeadObject operation: "
        r"No response body\.(.*))|"
        r"(?:(.*)AWS Error ACCESS_DENIED during HeadObject operation: No response "
        r"body\.(.*))$"
    )
    gcp_error_pattern = (
        r"^(?:(.*)google::cloud::Status\(UNAVAILABLE:(.*?)Couldn't resolve host name)"
    )
    if re.match(aws_error_pattern, str(error)):
        # Specially handle AWS error when reading files, to give a clearer error
        # message to avoid confusing users. The real issue is most likely that the AWS
        # S3 file credentials have not been properly configured yet.
        if isinstance(paths, str):
            # Quote to highlight single file path in error message for better
            # readability. List of file paths will be shown up as ['foo', 'boo'],
            # so only quote single file path here.
            paths = f'"{paths}"'
        raise OSError(
            (
                f"Failing to read AWS S3 file(s): {paths}. "
                "Please check that file exists and has properly configured access. "
                "You can also run AWS CLI command to get more detailed error message "
                "(e.g., aws s3 ls <file-name>). "
                "See https://awscli.amazonaws.com/v2/documentation/api/latest/reference/s3/index.html "  # noqa
                "and https://arrow.apache.org/docs/python/generated/pyarrow.fs.S3FileSystem.html "
                "for more information."
            )
        )
    elif re.match(gcp_error_pattern, str(error)):
        # Special handling for GCP errors (e.g., handling the special case of
        # requiring the filesystem to be instantiated with anonymous access to
        # read public files).
        if isinstance(paths, str):
            paths = f'"{paths}"'
        raise OSError(
            (
                f"Failing to read GCP GS file(s): {paths}. "
                "Please check that file exists and has properly configured access. "
                "If this is a public file, please instantiate a filesystem with "
                "anonymous access via `pyarrow.fs.GcsFileSystem(anonymous=True)` "
                "to read it. See https://google.aip.dev/auth/4110 and "
                "https://arrow.apache.org/docs/python/generated/pyarrow.fs.GcsFileSystem.html"  # noqa
                "for more information."
            )
        )

    else:
        raise error


def _is_local_windows_path(path: str) -> bool:
    """Determines if path is a Windows file-system location."""
    if sys.platform != "win32":
        return False

    if len(path) >= 1 and path[0] == "\\":
        return True
    if (
        len(path) >= 3
        and path[1] == ":"
        and (path[2] == "/" or path[2] == "\\")
        and path[0].isalpha()
    ):
        return True
    return False


def _unwrap_protocol(path):
    """
    Slice off any protocol prefixes on path.
    """
    if sys.platform == "win32" and _is_local_windows_path(path):
        # Represent as posix path such that downstream functions properly handle it.
        # This is executed when 'file://' is NOT included in the path.
        return pathlib.Path(path).as_posix()

    parsed = urllib.parse.urlparse(path, allow_fragments=False)  # support '#' in path
    query = "?" + parsed.query if parsed.query else ""  # support '?' in path
    netloc = parsed.netloc
    if parsed.scheme == "s3" and "@" in parsed.netloc:
        # If the path contains an @, it is assumed to be an anonymous
        # credentialed path, and we need to strip off the credentials.
        netloc = parsed.netloc.split("@")[-1]

    parsed_path = parsed.path
    # urlparse prepends the path with a '/'. This does not work on Windows
    # so if this is the case strip the leading slash.
    if (
        sys.platform == "win32"
        and not netloc
        and len(parsed_path) >= 3
        and parsed_path[0] == "/"  # The problematic leading slash
        and parsed_path[1].isalpha()  # Ensure it is a drive letter.
        and parsed_path[2:4] in (":", ":/")
    ):
        parsed_path = parsed_path[1:]

    return netloc + parsed_path + query


def _encode_url(path):
    return urllib.parse.quote(path, safe="/:")


def _decode_url(path):
    return urllib.parse.unquote(path)


def _resolve_custom_scheme(path: str) -> str:
    """Returns the resolved path if the given path follows a Ray-specific custom
    scheme. Othewise, returns the path unchanged.

    The supported custom schemes are: "local", "example".
    """
    parsed_uri = urllib.parse.urlparse(path)
    if parsed_uri.scheme == _LOCAL_SCHEME:
        path = parsed_uri.netloc + parsed_uri.path
    return path



import pytest
import pyarrow as pa

from deltacat.storage.model.expression import (
    Reference,
    Literal,
    Equal,
    NotEqual,
    GreaterThan,
    LessThan,
    GreaterThanEqual,
    LessThanEqual,
    And,
    Or,
    Not,
    In,
    Between,
    Like,
    IsNull,
)
from deltacat.storage.model.expression.visitor import DisplayVisitor, ExpressionVisitor


@pytest.fixture
def field_ref():
    return Reference("field1")


@pytest.fixture
def field_ref2():
    return Reference("field2")


@pytest.fixture
def literal_int():
    return Literal(pa.scalar(42))


@pytest.fixture
def literal_str():
    return Literal(pa.scalar("test"))


@pytest.fixture
def display_visitor():
    return DisplayVisitor()


class TestExpressionLibrary:
    """Test suite for the Deltacat expression library."""

    def test_reference_creation(self):
        ref = Reference("field1")
        assert ref.field == "field1"
        assert ref.index is None

    def test_reference_with_index(self):
        ref = Reference("field1", 0)
        assert ref.field == "field1"
        assert ref.index == 0

    def test_literal_creation(self):
        lit = Literal(pa.scalar(42))
        assert lit.value.as_py() == 42

    # Test the factory methods (.of)
    def test_factory_methods(self):
        # Reference.of
        ref = Reference.of("field1")
        assert ref.field == "field1"

        # Literal.of
        lit = Literal.of(42)
        assert lit.value.as_py() == 42

        # Equal.of with mixed types
        eq = Equal.of("field1", 42)
        assert isinstance(eq.left, Literal)
        assert isinstance(eq.right, Literal)
        assert eq.left.value.as_py() == "field1"
        assert eq.right.value.as_py() == 42

        # Not.of
        not_expr = Not.of(Equal.of("field1", 42))
        assert isinstance(not_expr.operand, Equal)

        # In.of
        in_expr = In.of("field1", [1, 2, 3])
        assert isinstance(in_expr.value, Literal)
        assert len(in_expr.values) == 3
        assert all(isinstance(v, Literal) for v in in_expr.values)

        # Between.of
        between_expr = Between.of("field1", 10, 20)
        assert isinstance(between_expr.value, Literal)
        assert between_expr.lower.value.as_py() == 10
        assert between_expr.upper.value.as_py() == 20

        # Like.of
        like_expr = Like.of("field1", "%test%")
        assert isinstance(like_expr.value, Literal)
        assert like_expr.pattern.value.as_py() == "%test%"

    # Test reference comparison helper methods
    def test_reference_comparison_helpers(self, field_ref):
        # Test eq, ne, gt, lt, ge, le methods
        eq_expr = field_ref.eq(42)
        assert isinstance(eq_expr, Equal)
        assert eq_expr.left == field_ref
        assert eq_expr.right.value.as_py() == 42

        ne_expr = field_ref.ne(42)
        assert isinstance(ne_expr, NotEqual)

        gt_expr = field_ref.gt(42)
        assert isinstance(gt_expr, GreaterThan)

        lt_expr = field_ref.lt(42)
        assert isinstance(lt_expr, LessThan)

        ge_expr = field_ref.ge(42)
        assert isinstance(ge_expr, GreaterThanEqual)

        le_expr = field_ref.le(42)
        assert isinstance(le_expr, LessThanEqual)

    # Test reference special operation helpers
    def test_reference_special_helpers(self, field_ref):
        # Test is_null, in_, between, like methods
        is_null_expr = field_ref.is_null()
        assert isinstance(is_null_expr, IsNull)
        assert is_null_expr.operand == field_ref

        in_expr = field_ref.in_([1, 2, 3])
        assert isinstance(in_expr, In)
        assert in_expr.value == field_ref
        assert len(in_expr.values) == 3
        assert in_expr.values[0].value.as_py() == 1

        between_expr = field_ref.between(10, 20)
        assert isinstance(between_expr, Between)
        assert between_expr.value == field_ref
        assert between_expr.lower.value.as_py() == 10
        assert between_expr.upper.value.as_py() == 20

        like_expr = field_ref.like("%test%")
        assert isinstance(like_expr, Like)
        assert like_expr.value == field_ref
        assert like_expr.pattern.value.as_py() == "%test%"

    # Test boolean expression helper methods
    def test_boolean_expression_helpers(self, field_ref):
        # Test and_, or_, not_ methods
        expr1 = field_ref.eq(42)
        expr2 = field_ref.gt(10)

        and_expr = expr1.and_(expr2)
        assert isinstance(and_expr, And)
        assert and_expr.left == expr1
        assert and_expr.right == expr2

        or_expr = expr1.or_(expr2)
        assert isinstance(or_expr, Or)
        assert or_expr.left == expr1
        assert or_expr.right == expr2

        not_expr = expr1.not_()
        assert isinstance(not_expr, Not)
        assert not_expr.operand == expr1

    # Test building complex expressions
    def test_complex_expression_building(self, field_ref, field_ref2):
        # Test building more complex expressions using method chaining
        expr = field_ref.eq(42).and_(field_ref2.gt(10)).or_(field_ref.is_null()).not_()

        assert isinstance(expr, Not)
        assert isinstance(expr.operand, Or)
        assert isinstance(expr.operand.left, And)
        assert isinstance(expr.operand.right, IsNull)

    # Test DisplayVisitor for different expression types
    def test_reference_display(self, field_ref, display_visitor):
        assert display_visitor.visit(field_ref) == "field1"

    def test_literal_display(self, literal_int, literal_str, display_visitor):
        assert display_visitor.visit(literal_int) == "42"
        assert display_visitor.visit(literal_str) == "test"

    def test_comparison_display(self, field_ref, literal_int, display_visitor):
        assert display_visitor.visit(Equal(field_ref, literal_int)) == "field1 = 42"
        assert display_visitor.visit(NotEqual(field_ref, literal_int)) == "field1 <> 42"
        assert (
            display_visitor.visit(GreaterThan(field_ref, literal_int)) == "field1 > 42"
        )
        assert display_visitor.visit(LessThan(field_ref, literal_int)) == "field1 < 42"
        assert (
            display_visitor.visit(GreaterThanEqual(field_ref, literal_int))
            == "field1 >= 42"
        )
        assert (
            display_visitor.visit(LessThanEqual(field_ref, literal_int))
            == "field1 <= 42"
        )

    def test_logical_operator_display(self, field_ref, literal_int, display_visitor):
        eq_expr = Equal(field_ref, literal_int)
        gt_expr = GreaterThan(field_ref, literal_int)

        assert (
            display_visitor.visit(And(eq_expr, gt_expr))
            == "(field1 = 42 AND field1 > 42)"
        )
        assert (
            display_visitor.visit(Or(eq_expr, gt_expr))
            == "(field1 = 42 OR field1 > 42)"
        )
        assert display_visitor.visit(Not(eq_expr)) == "NOT (field1 = 42)"

    def test_special_operator_display(self, field_ref, display_visitor):
        assert display_visitor.visit(IsNull(field_ref)) == "(field1) IS NULL"

        values = [Literal(pa.scalar(1)), Literal(pa.scalar(2)), Literal(pa.scalar(3))]
        assert display_visitor.visit(In(field_ref, values)) == "field1 IN (1, 2, 3)"

        lower = Literal(pa.scalar(10))
        upper = Literal(pa.scalar(20))
        assert (
            display_visitor.visit(Between(field_ref, lower, upper))
            == "field1 BETWEEN 10 AND 20"
        )

        pattern = Literal(pa.scalar("%test%"))
        assert display_visitor.visit(Like(field_ref, pattern)) == "field1 LIKE %test%"

    def test_complex_expression_display(self, field_ref, field_ref2, display_visitor):
        expr = field_ref.eq(42).and_(field_ref2.gt(10)).or_(field_ref.is_null()).not_()

        # Check that the DisplayVisitor correctly formats the complex expression
        assert (
            display_visitor.visit(expr)
            == "NOT (((field1 = 42 AND field2 > 10) OR (field1) IS NULL))"
        )

    # Test BinaryExpression with_ methods
    def test_binary_expression_with_methods(self, field_ref, field_ref2, literal_int):
        eq_expr = Equal(field_ref, literal_int)

        # Test with_left
        new_expr = eq_expr.with_left(field_ref2)
        assert isinstance(new_expr, Equal)
        assert new_expr.left == field_ref2
        assert new_expr.right == literal_int

        # Test with_right
        new_lit = Literal(pa.scalar(100))
        new_expr = eq_expr.with_right(new_lit)
        assert new_expr.left == field_ref
        assert new_expr.right == new_lit

    # Test __str__ method which uses DisplayVisitor
    def test_expression_str_method(self, field_ref, literal_int):
        eq_expr = Equal(field_ref, literal_int)
        assert str(eq_expr) == "field1 = 42"

    # Test proper parenthesization in complex expressions
    def test_nested_parentheses(self, field_ref, field_ref2, display_visitor):
        # Create a complex expression: (field1 = 1 AND field2 = 2) OR field2 = 3
        expr1 = Equal(field_ref, Literal(pa.scalar(1)))
        expr2 = Equal(field_ref2, Literal(pa.scalar(2)))
        expr3 = Equal(field_ref2, Literal(pa.scalar(3)))

        and_expr = And(expr1, expr2)
        or_expr = Or(and_expr, expr3)

        assert (
            display_visitor.visit(or_expr)
            == "((field1 = 1 AND field2 = 2) OR field2 = 3)"
        )

    # Test Literal comparison methods
    def test_literal_comparison_methods(self, literal_int):
        eq_expr = literal_int.eq("test")
        assert isinstance(eq_expr, Equal)
        assert eq_expr.left == literal_int
        assert eq_expr.right.value.as_py() == "test"

        ne_expr = literal_int.ne("test")
        assert isinstance(ne_expr, NotEqual)
        assert ne_expr.left == literal_int
        assert ne_expr.right.value.as_py() == "test"

    # Test a custom ExpressionVisitor implementation
    def test_custom_visitor(self, field_ref, literal_int):
        class CountingVisitor(ExpressionVisitor[None, int]):
            """Simple visitor that counts expression nodes"""

            def visit_reference(self, expr, context=None):
                return 1

            def visit_literal(self, expr, context=None):
                return 1

            def visit_binary_expression(self, expr, left, right, context=None):
                return left + right + 1

            def visit_unary_expression(self, expr, operand, context=None):
                return operand + 1

            def visit_in(self, expr, context=None):
                return 1 + len(expr.values) + 1  # value + all values + In operator

            def visit_between(self, expr, context=None):
                return 3  # value + lower + upper

            def visit_like(self, expr, context=None):
                return 2  # value + pattern

        visitor = CountingVisitor()

        # Count nodes in simple expressions
        assert visitor.visit(field_ref) == 1
        assert visitor.visit(literal_int) == 1
        assert visitor.visit(Equal(field_ref, literal_int)) == 3  # left + right + Equal

        # Count nodes in a more complex expression
        expr = field_ref.eq(42).and_(field_ref.gt(10))
        assert visitor.visit(expr) == 7  # (1+1+1) + (1+1+1) + 1



import posixpath
import pyarrow.fs

from deltacat.constants import REV_DIR_NAME
from deltacat.storage.model.partition import PartitionLocator
from deltacat.utils.filesystem import resolve_path_and_filesystem

"""
Helper functions to work with deltacat metadata paths.
TODO: Replace with direct calls to Deltacat storage interface.
"""


def _find_first_child_with_rev(
    parent_path: str, filesystem: pyarrow.fs.FileSystem
) -> str:
    """
    Walks the filesystem to find the first child directory with a `rev/` folder.

    This is a temporary solution to locate the first Namespace and Table directories.
    The Deltacat Storage interface will provide a more robust way to locate these directories.

    param: parent_path: The parent directory to search for a child with a `rev/` folder.
    param: filesystem: The filesystem to search for the child directory.
    returns: The name of the first child directory with a `rev/` folder.
    """
    children = filesystem.get_file_info(
        pyarrow.fs.FileSelector(parent_path, allow_not_found=True)
    )
    for child in children:
        if child.type == pyarrow.fs.FileType.Directory:
            rev_path = posixpath.join(child.path, REV_DIR_NAME)
            if filesystem.get_file_info(rev_path).type == pyarrow.fs.FileType.Directory:
                return child.base_name
    raise ValueError(f"No directory with 'rev/' found under {parent_path}")


def _find_table_path(root_path: str, filesystem: pyarrow.fs.FileSystem):
    """
    Finds a path with structure: root/namespace_id/table_id
    Uses _find_first_child_with_rev to determine the namespace and table ids.

    param: root_path: The root directory to search for the namespace and table directories.
    param: filesystem: The filesystem to search for the namespace and table directories.
    returns: The path to the table directory.
    raises: ValueError if the namespace or table directories are not found.
    """
    try:
        # Find Namespace (first directory under root with rev/)
        namespace_id = _find_first_child_with_rev(root_path, filesystem)
        namespace_path = posixpath.join(root_path, namespace_id)

        # Find Table (first directory under namespace with rev/)
        table_id = _find_first_child_with_rev(namespace_path, filesystem)
        return posixpath.join(namespace_path, table_id)

    except ValueError as e:
        raise ValueError(f"Failed to locate Namespace or Table: {e}") from e


def _find_partition_path(root_path: str, locator: PartitionLocator) -> str:
    """
    Finds the path to the partition directory for the specified locator.

    param: root_uri: The root URI of the dataset.
    param: locator: The DeltaLocator for the delta.
    returns: The path to the delta directory.
    """
    root_path, filesystem = resolve_path_and_filesystem(root_path)
    return posixpath.join(
        _find_table_path(root_path, filesystem),
        locator.table_version,
        locator.stream_id,
    )



from deltacat.experimental.storage.rivulet.schema.schema import Schema
from deltacat.experimental.storage.rivulet.schema.schema import Field
from deltacat.experimental.storage.rivulet.dataset import Dataset
from deltacat.experimental.storage.rivulet.schema.schema import Datatype

__all__ = [
    "Schema",
    "Field",
    "Dataset",
    "Datatype",
]
